#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from oslo_serialization import jsonutils

from tempest.lib.common.utils import data_utils

from manilaclient.tests.functional.osc import base


class SharesGroupTypesCLITest(base.OSCClientTestBase):

    def test_share_group_type_create(self):
        share_group_type_name = 'test_share_group_type_create'
        share_group_type = self.create_share_group_type(
            name=share_group_type_name, share_types='dhss_false')

        self.assertEqual(share_group_type['name'], share_group_type_name)
        self.assertIsNotNone(share_group_type['share_types'])

        shares_group_type_list = self.listing_result(
            'share', 'group type list')
        self.assertIn(
            share_group_type['id'],
            [item['ID'] for item in shares_group_type_list])

    def test_openstack_share_group_type_list(self):
        share_group_type_name = data_utils.rand_name(
            'test_share_group_type_create')
        share_group_type = self.create_share_group_type(
            name=share_group_type_name, share_types='dhss_false')

        shares_group_type_list = self.listing_result(
            'share', 'group type list')
        self.assertTableStruct(shares_group_type_list, [
            'ID',
            'Name',
            'Share Types',
            'Visibility',
            'Is Default',
            'Group Specs'
        ])
        self.assertIn(
            share_group_type['id'],
            [item['ID'] for item in shares_group_type_list])

    def test_openstack_share_group_type_show(self):
        share_group_type_name = data_utils.rand_name(
            'test_share_group_type_create')
        share_type_name = 'dhss_false'
        share_group_type = self.create_share_group_type(
            name=share_group_type_name, share_types=share_type_name)

        shares_group_type_show = self.dict_result(
            'share', f'group type show {share_group_type_name}')
        share_type_id = self.dict_result(
            'share', f'type show {share_type_name}')['id']

        expected_sgt_values = {
            'id': share_group_type['id'],
            'name': share_group_type_name,
            'share_types': share_type_id,
            'visibility': 'public',
            'is_default': 'False',
            'group_specs': ''
        }

        for k, v in shares_group_type_show.items():
            self.assertEqual(expected_sgt_values[k], shares_group_type_show[k])

    def test_openstack_share_group_type_set(self):
        share_group_type_name = data_utils.rand_name(
            'test_share_group_type_create')
        share_type_name = 'dhss_false'
        share_group_type = self.create_share_group_type(
            name=share_group_type_name, share_types=share_type_name)
        shares_group_type_show = self.openstack(
            f'share group type show {share_group_type_name} -f json')
        shares_group_type_show = jsonutils.loads(shares_group_type_show)

        expected_sgt_values = {
            'id': share_group_type['id'],
            'group_specs': {}
        }

        for k, v in expected_sgt_values.items():
            self.assertEqual(expected_sgt_values[k], shares_group_type_show[k]
                             )

        group_snap_key = 'snapshot_support'
        group_snap_value = 'False'
        group_specs = f"{group_snap_key}={group_snap_value}"

        self.dict_result(
            'share',
            f'group type set {share_group_type_name} '
            f'--group-specs {group_specs}')

        shares_group_type_show = self.openstack(
            f'share group type show {share_group_type_name} -f json')
        shares_group_type_show = jsonutils.loads(shares_group_type_show)

        expected_sgt_values = {
            'id': share_group_type['id'],
            'group_specs': {
                group_snap_key: group_snap_value
            }
        }

        for k, v in expected_sgt_values.items():
            self.assertEqual(expected_sgt_values[k], shares_group_type_show[k])

    def test_openstack_share_group_type_unset(self):
        share_group_type_name = data_utils.rand_name(
            'test_share_group_type_create')
        group_snap_key = 'snapshot_support'
        group_snap_value = 'False'
        group_specs = f"{group_snap_key}={group_snap_value}"
        share_group_type = self.create_share_group_type(
            name=share_group_type_name, share_types='dhss_false',
            group_specs=group_specs)
        shares_group_type_show = self.openstack(
            f'share group type show {share_group_type_name} -f json')
        shares_group_type_show = jsonutils.loads(shares_group_type_show)

        expected_sgt_values = {
            'id': share_group_type['id'],
            'group_specs': {
                group_snap_key: group_snap_value
            }
        }

        for k, v in expected_sgt_values.items():
            self.assertEqual(expected_sgt_values[k], shares_group_type_show[k]
                             )

        self.dict_result(
            'share',
            f'group type unset {share_group_type_name} '
            f'{group_snap_key}')

        shares_group_type_show = self.openstack(
            f'share group type show {share_group_type_name} -f json')
        shares_group_type_show = jsonutils.loads(shares_group_type_show)

        expected_sgt_values = {
            'id': share_group_type['id'],
            'group_specs': {}
        }

        for k, v in expected_sgt_values.items():
            self.assertEqual(expected_sgt_values[k], shares_group_type_show[k])



# Copyright 2010 Jacob Kaplan-Moss
# Copyright 2011 OpenStack Foundation
# Copyright 2011 Piston Cloud Computing, Inc.
# Copyright 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
from urllib import parse

from oslo_serialization import jsonutils
from oslo_utils import importutils
from oslo_utils import strutils
import re
import requests

from manilaclient import exceptions

try:
    from eventlet import sleep
except ImportError:
    from time import sleep  # noqa

try:
    osprofiler_web = importutils.try_import("osprofiler.web")
except Exception:
    pass


class HTTPClient(object):
    """HTTP Client class used by multiple clients.

    The imported Requests module caches and reuses objects with the same
    destination. To avoid the problem of sending duplicate requests it is
    necessary that the Requests module is only imported once during client
    execution. This class is shared by multiple client versions so that the
    client can be changed to another version during execution.
    """
    API_VERSION_HEADER = "X-Openstack-Manila-Api-Version"

    def __init__(self, endpoint_url, token, user_agent, api_version,
                 insecure=False, cacert=None, timeout=None, retries=None,
                 http_log_debug=False, cert=None):
        self.endpoint_url = endpoint_url
        self.base_url = self._get_base_url(self.endpoint_url)
        self.retries = int(retries or 0)
        self.http_log_debug = http_log_debug

        self.request_options = self._set_request_options(
            insecure, cacert, timeout, cert)

        self.default_headers = {
            'X-Auth-Token': token,
            self.API_VERSION_HEADER: api_version.get_string(),
            'User-Agent': user_agent,
            'Accept': 'application/json',
        }

        self._add_log_handlers(http_log_debug)

    def _add_log_handlers(self, http_log_debug):
        self._logger = logging.getLogger(__name__)

        # check that handler hasn't already been added
        if http_log_debug and not self._logger.handlers:
            ch = logging.StreamHandler()
            ch._name = 'http_client_handler'
            self._logger.setLevel(logging.DEBUG)
            self._logger.addHandler(ch)

            if hasattr(requests, 'logging'):
                rql = requests.logging.getLogger(requests.__name__)
                rql.addHandler(ch)

    def _get_base_url(self, url):
        """Truncates url and returns base endpoint"""
        service_endpoint = parse.urlparse(url)
        service_endpoint_base_path = re.search(
            r'(.+?)/v([0-9]+|[0-9]+\.[0-9]+)(/.*|$)', service_endpoint.path)
        base_path = (service_endpoint_base_path.group(1)
                     if service_endpoint_base_path else '')
        base_url = service_endpoint._replace(path=base_path)
        return parse.urlunparse(base_url) + '/'

    def _set_request_options(self, insecure, cacert, timeout=None, cert=None):
        options = {'verify': True}

        if insecure:
            options['verify'] = False
        elif cacert:
            options['verify'] = cacert
        if cert:
            options['cert'] = cert

        if timeout:
            options['timeout'] = timeout

        return options

    def request(self, url, method, **kwargs):
        headers = copy.deepcopy(self.default_headers)
        headers.update(kwargs.get('headers', {}))

        options = copy.deepcopy(self.request_options)

        if osprofiler_web:
            headers.update(osprofiler_web.get_trace_id_headers())

        if 'body' in kwargs:
            headers['Content-Type'] = 'application/json'
            options['data'] = jsonutils.dumps(kwargs['body'])

        self.log_request(method, url, headers, options.get('data', None))
        resp = requests.request(method, url, headers=headers, **options)
        self.log_response(resp)

        body = None

        if resp.text:
            try:
                body = jsonutils.loads(resp.text)
            except ValueError:
                pass

        if resp.status_code >= 400:
            raise exceptions.from_response(resp, method, url)

        return resp, body

    def _cs_request(self, url, method, **kwargs):
        return self._cs_request_with_retries(
            self.endpoint_url + url,
            method,
            **kwargs)

    def _cs_request_base_url(self, url, method, **kwargs):
        return self._cs_request_with_retries(
            self.base_url + url,
            method,
            **kwargs)

    def _cs_request_with_retries(self, url, method, **kwargs):
        attempts = 0
        timeout = 1
        while True:
            attempts += 1
            try:
                resp, body = self.request(url, method, **kwargs)
                return resp, body
            except (exceptions.BadRequest,
                    requests.exceptions.RequestException,
                    exceptions.ClientException) as e:
                if attempts > self.retries:
                    raise

                self._logger.debug("Request error: %s", str(e))

            self._logger.debug(
                "Failed attempt(%(current)s of %(total)s), "
                " retrying in %(sec)s seconds", {
                    'current': attempts,
                    'total': self.retries,
                    'sec': timeout
                })
            sleep(timeout)
            timeout *= 2

    def get_with_base_url(self, url, **kwargs):
        return self._cs_request_base_url(url, 'GET', **kwargs)

    def get(self, url, **kwargs):
        return self._cs_request(url, 'GET', **kwargs)

    def post(self, url, **kwargs):
        return self._cs_request(url, 'POST', **kwargs)

    def put(self, url, **kwargs):
        return self._cs_request(url, 'PUT', **kwargs)

    def delete(self, url, **kwargs):
        return self._cs_request(url, 'DELETE', **kwargs)

    def log_request(self, method, url, headers, data=None):
        if not self.http_log_debug:
            return

        string_parts = ['curl -i', ' -X %s' % method, ' %s' % url]

        for element in headers:
            header = ' -H "%s: %s"' % (element, headers[element])
            string_parts.append(header)

        if data:
            if "password" in data:
                data = strutils.mask_password(data)
            string_parts.append(" -d '%s'" % data)
        self._logger.debug("\nREQ: %s\n", "".join(string_parts))

    def log_response(self, resp):
        if not self.http_log_debug:
            return
        self._logger.debug(
            "RESP: [%(code)s] %(headers)s\nRESP BODY: %(body)s\n", {
                'code': resp.status_code,
                'headers': resp.headers,
                'body': resp.text
            })



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.common.apiclient import utils as apiutils


LOG = logging.getLogger(__name__)

MESSAGE_ATTRIBUTES = [
    'id',
    'resource_type',
    'resource_id',
    'action_id',
    'user_message',
    'message_level',
    'detail_id',
    'created_at',
    'expires_at',
    'request_id'
]


class DeleteMessage(command.Command):
    """Remove one or more messages."""
    _description = _("Remove one or more messages")

    def get_parser(self, prog_name):
        parser = super(DeleteMessage, self).get_parser(prog_name)
        parser.add_argument(
            'message',
            metavar='<message>',
            nargs='+',
            help=_('ID of the message(s).'))
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        failure_count = 0

        for message in parsed_args.message:
            try:
                message_ref = apiutils.find_resource(
                    share_client.messages,
                    message)
                share_client.messages.delete(message_ref)
            except Exception as e:
                failure_count += 1
                LOG.error(_(
                    "Delete for message %(message)s failed: %(e)s"),
                    {'message': message, 'e': e})

        if failure_count > 0:
            raise exceptions.CommandError(_(
                "Unable to delete some or all of the specified messages."))


class ListMessage(command.Lister):
    """Lists all messages."""
    _description = _("Lists all messages")

    def get_parser(self, prog_name):
        parser = super(ListMessage, self).get_parser(prog_name)
        parser.add_argument(
            '--resource-id',
            metavar='<resource-id>',
            default=None,
            help=_('Filters results by a resource uuid. Default=None.'))
        parser.add_argument(
            '--resource-type',
            metavar='<resource-type>',
            default=None,
            help=_('Filters results by a resource type. Default=None. '
                   'Example: "openstack message list --resource-type share"'))
        parser.add_argument(
            '--action-id',
            metavar='<action-id>',
            default=None,
            help=_('Filters results by action id. Default=None.'))
        parser.add_argument(
            '--detail-id',
            metavar='<detail-id>',
            default=None,
            help=_('Filters results by detail id. Default=None.'))
        parser.add_argument(
            '--request-id',
            metavar='<request-id>',
            default=None,
            help=_('Filters results by request id. Default=None.'))
        parser.add_argument(
            '--message-level',
            metavar='<message-level>',
            default=None,
            help=_('Filters results by the message level. Default=None. '
                   'Example: "openstack message list --message-level ERROR".'))
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            default=None,
            help=_('Maximum number of messages to return. (Default=None)'))
        parser.add_argument(
            '--since',
            metavar='<since>',
            default=None,
            help=_('Return only user messages created since given date. '
                   'The date format must be conforming to ISO8601. '
                   'Available only for microversion >= 2.52.'))
        parser.add_argument(
            '--before',
            metavar='<before>',
            default=None,
            help=_('Return only user messages created before given date. '
                   'The date format must be conforming to ISO8601. '
                   'Available only for microversion >= 2.52.'))
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        search_opts = {
            'limit': parsed_args.limit,
            'request_id': parsed_args.request_id,
            'resource_type': parsed_args.resource_type,
            'resource_id': parsed_args.resource_id,
            'action_id': parsed_args.action_id,
            'detail_id': parsed_args.detail_id,
            'message_level': parsed_args.message_level
        }

        if share_client.api_version < api_versions.APIVersion("2.52"):
            if getattr(parsed_args, 'since') or getattr(parsed_args, 'before'):
                raise exceptions.CommandError(_(
                    "Filtering messages by 'since' and 'before'"
                    " is possible only with Manila API version >=2.52"))
        else:
            search_opts['created_since'] = parsed_args.since
            search_opts['created_before'] = parsed_args.before

        messages = share_client.messages.list(search_opts=search_opts)
        columns = [
            'ID',
            'Resource Type',
            'Resource ID',
            'Action ID',
            'User Message',
            'Detail ID',
            'Created At']

        return (columns, (oscutils.get_item_properties
                (m, columns) for m in messages))


class ShowMessage(command.ShowOne):
    """Show details about a message."""
    _description = _("Show details about a message")

    def get_parser(self, prog_name):
        parser = super(ShowMessage, self).get_parser(prog_name)
        parser.add_argument(
            'message',
            metavar='<message>',
            help=_('ID of the message.'))
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        message = apiutils.find_resource(
            share_client.messages,
            parsed_args.message)

        return (MESSAGE_ATTRIBUTES, oscutils.get_dict_properties(
            message._info, MESSAGE_ATTRIBUTES))



# Copyright 2019 NetApp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from manilaclient.tests.functional import base
from manilaclient.tests.functional import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions


@ddt.ddt
@utils.skip_if_microversion_not_supported('2.51')
class ShareNetworkSubnetsReadWriteTest(base.BaseTestCase):

    def setUp(self):
        super(ShareNetworkSubnetsReadWriteTest, self).setUp()
        self.name = data_utils.rand_name('autotest')
        self.description = 'fake_description'
        self.neutron_net_id = 'fake_neutron_net_id'
        self.neutron_subnet_id = 'fake_neutron_subnet_id'

        self.sn = self.create_share_network(
            name=self.name,
            description=self.description,
            neutron_net_id=self.neutron_net_id,
            neutron_subnet_id=self.neutron_subnet_id,
        )

    def test_get_share_network_subnet(self):
        default_subnet = utils.get_default_subnet(self.user_client,
                                                  self.sn['id'])

        subnet = self.user_client.get_share_network_subnet(
            self.sn['id'], default_subnet['id'])

        self.assertEqual(self.neutron_net_id, subnet['neutron_net_id'])
        self.assertEqual(self.neutron_subnet_id, subnet['neutron_subnet_id'])

    def test_get_invalid_share_network_subnet(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.user_client.get_share_network_subnet,
            self.sn['id'], 'invalid_subnet_id')

    def _get_availability_zone(self):
        availability_zones = self.user_client.list_availability_zones()
        return availability_zones[0]['Name']

    def test_add_share_network_subnet_to_share_network(self):
        neutron_net_id = 'new_neutron_net_id'
        neutron_subnet_id = 'new_neutron_subnet_id'
        availability_zone = self._get_availability_zone()

        subnet = self.add_share_network_subnet(
            self.sn['id'],
            neutron_net_id, neutron_subnet_id,
            availability_zone,
            cleanup_in_class=False)

        self.assertEqual(neutron_net_id, subnet['neutron_net_id'])
        self.assertEqual(neutron_subnet_id, subnet['neutron_subnet_id'])
        self.assertEqual(availability_zone, subnet['availability_zone'])

    @ddt.data(
        {'neutron_net_id': None, 'neutron_subnet_id': 'fake_subnet_id'},
        {'neutron_net_id': 'fake_net_id', 'neutron_subnet_id': None},
        {'availability_zone': 'invalid_availability_zone'},
    )
    def test_add_invalid_share_network_subnet_to_share_network(self, params):
        self.assertRaises(
            exceptions.CommandFailed,
            self.add_share_network_subnet,
            self.sn['id'],
            **params)

    def test_add_share_network_subnet_to_invalid_share_network(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.add_share_network_subnet,
            'invalid_share_network',
            self.neutron_net_id,
            self.neutron_subnet_id)

    def test_add_delete_share_network_subnet_from_share_network(self):
        neutron_net_id = 'new_neutron_net_id'
        neutron_subnet_id = 'new_neutron_subnet_id'
        availability_zone = self._get_availability_zone()

        subnet = self.add_share_network_subnet(
            self.sn['id'],
            neutron_net_id, neutron_subnet_id,
            availability_zone,
            cleanup_in_class=False)
        self.user_client.delete_share_network_subnet(
            share_network_subnet=subnet['id'],
            share_network=self.sn['id'])

        self.user_client.wait_for_share_network_subnet_deletion(
            share_network_subnet=subnet['id'],
            share_network=self.sn['id'])

    def test_delete_invalid_share_network_subnet(self):
        self.assertRaises(
            exceptions.NotFound,
            self.user_client.delete_share_network_subnet,
            share_network_subnet='invalid_subnet_id',
            share_network=self.sn['id'])



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.


from manilaclient.tests.functional.osc import base


class AvailabilityZonesCLITest(base.OSCClientTestBase):

    def test_openstack_share_availability_zones_list(self):
        azs = self.listing_result('share', 'availability zone list')
        self.assertTableStruct(azs, [
            'Id',
            'Name',
            'Created At',
            'Updated At'
        ])
        self.assertTrue(len(azs) > 0)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from operator import xor

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.common._i18n import _


LOG = logging.getLogger(__name__)


class CreateShareNetworkSubnet(command.ShowOne):
    """Create a share network subnet."""
    _description = _("Create a share network subnet")

    def get_parser(self, prog_name):
        parser = super(CreateShareNetworkSubnet, self).get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Share network name or ID.")
        )
        parser.add_argument(
            "--neutron-net-id",
            metavar="<neutron-net-id>",
            default=None,
            help=_("Neutron network ID. Used to set up network for share "
                   "servers (optional). Should be defined together with "
                   "'--neutron-subnet-id'.")
        )
        parser.add_argument(
            "--neutron-subnet-id",
            metavar="<neutron-subnet-id>",
            default=None,
            help=_("Neutron subnet ID. Used to set up network for share "
                   "servers (optional). Should be defined together with "
                   "'--neutron-net-id' to which this subnet belongs to. ")
        )
        parser.add_argument(
            "--availability-zone",
            metavar="<availability-zone>",
            default=None,
            help=_("Optional availability zone that the subnet is available "
                   "within (Default=None). If None, the subnet will be "
                   "considered as being available across all availability "
                   "zones.")
        )
        parser.add_argument(
            '--check-only',
            default=False,
            action='store_true',
            help=_("Run a dry-run of a share network subnet create. "
                   "Available only for microversion >= 2.70.")
        )
        parser.add_argument(
            '--restart-check',
            default=False,
            action='store_true',
            help=_("Restart a dry-run of a share network subnet create. "
                   "Helpful when check results are stale. "
                   "Available only for microversion >= 2.70.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        # check and restart check during create is only available from 2.70.
        if (parsed_args.check_only and
                share_client.api_version < api_versions.APIVersion("2.70")):
            raise exceptions.CommandError(
                "Check only can be specified only with manila API "
                "version >= 2.70.")
        if (parsed_args.restart_check and
                share_client.api_version < api_versions.APIVersion("2.70")):
            raise exceptions.CommandError(
                "Restart check can be specified only with manila API "
                "version >= 2.70.")

        if xor(bool(parsed_args.neutron_net_id),
               bool(parsed_args.neutron_subnet_id)):
            raise exceptions.CommandError(
                "Both neutron_net_id and neutron_subnet_id should be "
                "specified. Alternatively, neither of them should be "
                "specified.")

        share_network_id = oscutils.find_resource(
            share_client.share_networks,
            parsed_args.share_network).id

        if parsed_args.check_only or parsed_args.restart_check:
            subnet_create_check = (
                share_client.share_networks.share_network_subnet_create_check(
                    neutron_net_id=parsed_args.neutron_net_id,
                    neutron_subnet_id=parsed_args.neutron_subnet_id,
                    availability_zone=parsed_args.availability_zone,
                    reset_operation=parsed_args.restart_check,
                    share_network_id=share_network_id)
            )
            subnet_data = subnet_create_check[1]
        else:
            share_network_subnet = share_client.share_network_subnets.create(
                neutron_net_id=parsed_args.neutron_net_id,
                neutron_subnet_id=parsed_args.neutron_subnet_id,
                availability_zone=parsed_args.availability_zone,
                share_network_id=share_network_id
            )
            subnet_data = share_network_subnet._info

        return self.dict2columns(subnet_data)


class DeleteShareNetworkSubnet(command.Command):
    """Delete a share network subnet."""
    _description = _("Delete a share network subnet")

    def get_parser(self, prog_name):
        parser = super(DeleteShareNetworkSubnet, self).get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Share network name or ID.")
        )
        parser.add_argument(
            "share_network_subnet",
            metavar="<share-network-subnet>",
            nargs="+",
            help=_("ID(s) of share network subnet(s) to be deleted.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        share_network_id = oscutils.find_resource(
            share_client.share_networks,
            parsed_args.share_network).id

        for subnet in parsed_args.share_network_subnet:
            try:
                share_client.share_network_subnets.delete(
                    share_network_id,
                    subnet)

            except Exception as e:
                result += 1
                LOG.error(f"Failed to delete share network subnet with "
                          f"ID {subnet}: {e}")
        if result > 0:
            total = len(parsed_args.share_network_subnet)
            raise exceptions.CommandError(
                f"{result} of {total} share network subnets failed to be "
                f"deleted.")


class ShowShareNetworkSubnet(command.ShowOne):
    """Show share network subnet."""
    _description = _("Show share network subnet")

    def get_parser(self, prog_name):
        parser = super(ShowShareNetworkSubnet, self).get_parser(prog_name)
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Share network name or ID.")
        )
        parser.add_argument(
            "share_network_subnet",
            metavar="<share-network-subnet>",
            help=_("ID of share network subnet to show.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_network_id = oscutils.find_resource(
            share_client.share_networks,
            parsed_args.share_network).id

        share_network_subnet = share_client.share_network_subnets.get(
            share_network_id,
            parsed_args.share_network_subnet)

        return self.dict2columns(share_network_subnet._info)



# Copyright 2015 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from manilaclient import api_versions
from manilaclient import extension
from manilaclient.tests.unit import utils
from manilaclient.tests.unit.v2 import fakes
from manilaclient.v2 import share_instances


extensions = [
    extension.Extension('share_instances', share_instances),
]
cs = fakes.FakeClient(extensions=extensions)


@ddt.ddt
class ShareInstancesTest(utils.TestCase):

    def _get_manager(self, microversion):
        version = api_versions.APIVersion(microversion)
        mock_microversion = mock.Mock(api_version=version)
        return share_instances.ShareInstanceManager(api=mock_microversion)

    def test_list(self):
        cs.share_instances.list(search_opts=None)
        cs.assert_called('GET', '/share_instances')

    @ddt.data(('id', 'b4991315-eb7d-43ec-979e-5715d4399827'),
              ('path', '//0.0.0.0/fake_path'))
    @ddt.unpack
    def test_list_by_export_location(self, filter_type, value):
        cs.share_instances.list(export_location=value)
        cs.assert_called(
            'GET', '/share_instances?export_location_' +
            filter_type + '=' + value)

    def test_get(self):
        instance = type('None', (object, ), {'id': '1234'})
        cs.share_instances.get(instance)
        cs.assert_called('GET', '/share_instances/1234')

    @ddt.data(
        ("2.6", type("InstanceUUID", (object, ), {"uuid": "1234"})),
        ("2.7", type("InstanceUUID", (object, ), {"uuid": "1234"})),
        ("2.6", type("InstanceID", (object, ), {"id": "1234"})),
        ("2.7", type("InstanceID", (object, ), {"id": "1234"})),
        ("2.6", "1234"),
        ("2.7", "1234"),
    )
    @ddt.unpack
    def test_reset_instance_state(self, microversion, instance):
        manager = self._get_manager(microversion)
        state = 'available'
        if (api_versions.APIVersion(microversion) >
                api_versions.APIVersion("2.6")):
            action_name = "reset_status"
        else:
            action_name = "os-reset_status"

        with mock.patch.object(manager, "_action", mock.Mock()):
            manager.reset_state(instance, state)

            manager._action.assert_called_once_with(
                action_name, instance, {"status": state})

    @ddt.data(
        ("2.6", type('InstanceUUID', (object, ), {"uuid": "1234"})),
        ("2.6", "1234"),
        ("2.7", type('InstanceUUID', (object, ), {"uuid": "1234"})),
        ("2.7", "1234"),
    )
    @ddt.unpack
    def test_force_delete_share_snapshot(self, microversion, instance):
        manager = self._get_manager(microversion)
        if (api_versions.APIVersion(microversion) >
                api_versions.APIVersion("2.6")):
            action_name = "force_delete"
        else:
            action_name = "os-force_delete"

        with mock.patch.object(manager, "_action", mock.Mock()):
            manager.force_delete(instance)

            manager._action.assert_called_once_with(action_name, "1234")

    @ddt.data(
        ("2.6", "1234", "migrating_to"),
        ("2.6", "1234", "error"),
        ("2.6", "1234", "available"),
        ("2.7", "1234", "error_deleting"),
        ("2.7", "1234", "deleting"),
        ("2.7", "1234", "migrating"),
    )
    @ddt.unpack
    def test_valid_instance_state(self, microversion, instance, state):
        manager = self._get_manager(microversion)
        if (api_versions.APIVersion(microversion) >
                api_versions.APIVersion("2.6")):
            action_name = "reset_status"
        else:
            action_name = "os-reset_status"

        with mock.patch.object(manager, "_action", mock.Mock()):
            manager.reset_state(instance, state)

            manager._action.assert_called_once_with(
                action_name, instance, {"status": state})



# Copyright 2021 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from osc_lib.command import command
from osc_lib import utils as oscutils

from manilaclient.common._i18n import _


class ShareLimitsShow(command.Lister):
    """Show a list of share limits for a user."""

    _description = _("Show a list of share limits for a user.")

    def get_parser(self, prog_name):
        parser = super(ShareLimitsShow, self).get_parser(prog_name)
        limit_type_group = parser.add_mutually_exclusive_group(required=True)

        limit_type_group.add_argument(
            '--absolute',
            action='store_true',
            default=False,
            help=_('Get the absolute limits for the user')
        )
        limit_type_group.add_argument(
            '--rate',
            action='store_true',
            default=False,
            help=_('Get the API rate limits for the user')
        )

        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        # limit_type = 'absolute'
        if parsed_args.rate:
            # limit_type = 'rate'
            columns = [
                "Verb",
                "Regex",
                "URI",
                "Value",
                "Remaining",
                "Unit",
                "Next Available",
            ]
            data = list(share_client.limits.get().rate)
        else:
            columns = [
                'Name',
                'Value',
            ]

            data = list(share_client.limits.get().absolute)

        return (columns, (oscutils.get_item_properties(s, columns)
                          for s in data))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from osc_lib.command import command
from osc_lib import utils

from manilaclient.common._i18n import _
from manilaclient.common.apiclient import utils as apiutils


class ShareSnapshotInstanceExportLocationList(command.Lister):
    """List export locations from a share snapshot instance."""
    _description = _("List export locations from a share snapshot instance.")

    def get_parser(self, prog_name):
        parser = (
            super(ShareSnapshotInstanceExportLocationList, self).get_parser(
                prog_name))
        parser.add_argument(
            "instance",
            metavar="<instance>",
            help=_("Name or ID of the share instance.")
        )

        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        snapshot_instance = apiutils.find_resource(
            share_client.share_snapshot_instances, parsed_args.instance)

        share_snapshot_instance_export_locations = (
            share_client.share_snapshot_instance_export_locations.list(
                snapshot_instance=snapshot_instance))

        columns = ["ID", "Path", "Is Admin only"]

        return (
            columns,
            (utils.get_item_properties(s, columns) for s in
             share_snapshot_instance_export_locations))


class ShareSnapshotInstanceExportLocationShow(command.ShowOne):
    """Show export location of the share snapshot instance."""
    _description = _("Show export location of the share snapshot instance.")

    def get_parser(self, prog_name):
        parser = (
            super(ShareSnapshotInstanceExportLocationShow, self).get_parser(
                prog_name))
        parser.add_argument(
            'snapshot_instance',
            metavar='<snapshot_instance>',
            help=_('ID of the share snapshot instance.')
        )
        parser.add_argument(
            'export_location',
            metavar='<export_location>',
            help=_('ID of the share snapshot instance export location.')
        )

        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        snapshot_instance = apiutils.find_resource(
            share_client.share_snapshot_instances,
            parsed_args.snapshot_instance)

        share_snapshot_instance_export_location = (
            share_client.share_snapshot_instance_export_locations.get(
                parsed_args.export_location,
                snapshot_instance=snapshot_instance))

        share_snapshot_instance_export_location._info.pop('links', None)

        return self.dict2columns(share_snapshot_instance_export_location._info)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as oscutils
from oslo_utils import strutils

from manilaclient.common._i18n import _
from manilaclient.common.apiclient import utils as apiutils
from manilaclient.osc import utils

LOG = logging.getLogger(__name__)

ATTRIBUTES = [
    'id',
    'name',
    'share_types',
    'visibility',
    'is_default',
    'group_specs'
]


class CreateShareGroupType(command.ShowOne):
    """Create new share group type."""
    _description = _(
        "Create new share group type")

    log = logging.getLogger(__name__ + ".CreateShareGroupType")

    def get_parser(self, prog_name):
        parser = super(CreateShareGroupType, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar="<name>",
            default=None,
            help=_('Share group type name')
        )
        parser.add_argument(
            "share_types",
            metavar="<share-types>",
            nargs="+",
            default=None,
            help=_("List of share type names or IDs. Example:"
                   " my-share-type-1 my-share-type-2"),
        )
        parser.add_argument(
            "--group-specs",
            type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
            help=_("Share Group type extra specs by key and value."
                   " OPTIONAL: Default=None. Example:"
                   " --group-specs consistent_snapshot_support=host."),
        )
        parser.add_argument(
            '--public',
            metavar="<public>",
            default=True,
            help=_('Make type accessible to the public (default true).')
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        kwargs = {
            'name': parsed_args.name
        }

        share_types_list = []
        for share_type in parsed_args.share_types:
            try:
                share_type_obj = apiutils.find_resource(
                    share_client.share_types,
                    share_type)

                share_types_list.append(share_type_obj.name)
            except Exception as e:
                msg = LOG.error(_("Failed to find the share type with "
                                  "name or ID '%(share_type)s': %(e)s"),
                                {'share_type': share_type, 'e': e})
                raise exceptions.CommandError(msg)

        kwargs['share_types'] = share_types_list

        if parsed_args.public:
            kwargs['is_public'] = strutils.bool_from_string(
                parsed_args.public, default=True)

        group_specs = {}
        if parsed_args.group_specs:
            for item in parsed_args.group_specs:
                group_specs = utils.extract_group_specs(group_specs, [item])

        kwargs['group_specs'] = group_specs

        share_group_type = share_client.share_group_types.create(**kwargs)

        formatter = parsed_args.formatter

        formatted_group_type = utils.format_share_group_type(
            share_group_type, formatter)

        return (ATTRIBUTES, oscutils.get_dict_properties(
                formatted_group_type, ATTRIBUTES))


class DeleteShareGroupType(command.Command):
    """Delete a share group type."""
    _description = _("Delete a share group type")

    log = logging.getLogger(__name__ + ".DeleteShareGroupType")

    def get_parser(self, prog_name):
        parser = super(DeleteShareGroupType, self).get_parser(prog_name)
        parser.add_argument(
            'share_group_types',
            metavar="<share-group-types>",
            nargs="+",
            help=_("Name or ID of the share group type(s) to delete")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for share_group_type in parsed_args.share_group_types:
            try:
                share_group_type_obj = apiutils.find_resource(
                    share_client.share_group_types,
                    share_group_type)

                share_client.share_group_types.delete(share_group_type_obj)
            except Exception as e:
                result += 1
                LOG.error(_(
                    "Failed to delete share group type with "
                    "name or ID '%(share_group_type)s': %(e)s"),
                    {'share_group_type': share_group_type, 'e': e})

        if result > 0:
            total = len(parsed_args.share_group_types)
            msg = (_("%(result)s of %(total)s share group types failed "
                   "to delete.") % {'result': result, 'total': total})
            raise exceptions.CommandError(msg)


class ListShareGroupType(command.Lister):
    """List Share Group Types."""
    _description = _("List share types")

    log = logging.getLogger(__name__ + ".ListShareGroupType")

    def get_parser(self, prog_name):
        parser = super(ListShareGroupType, self).get_parser(prog_name)
        parser.add_argument(
            '--all',
            action='store_true',
            default=False,
            help=_('Display all share group types whether public or private. '
                   'Default=False. (Admin only)'),
        )
        parser.add_argument(
            '--group-specs',
            type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
            help=_('Filter share group types with group specs (key=value).'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        search_opts = {}
        if parsed_args.group_specs:
            search_opts = {
                'group_specs': utils.extract_group_specs(
                    extra_specs={},
                    specs_to_add=parsed_args.group_specs)
            }

        formatter = parsed_args.formatter

        share_group_types = share_client.share_group_types.list(
            search_opts=search_opts,
            show_all=parsed_args.all)

        formatted_types = []
        for share_group_type in share_group_types:
            formatted_types.append(utils.format_share_group_type(
                share_group_type, formatter))

        column_headers = utils.format_column_headers(ATTRIBUTES)
        values = (oscutils.get_dict_properties(
            sgt, ATTRIBUTES) for sgt in formatted_types)

        return (column_headers, values)


class ShowShareGroupType(command.ShowOne):
    """Show Share Group Types."""
    _description = _("Show share group types")

    log = logging.getLogger(__name__ + ".ShowShareGroupType")

    def get_parser(self, prog_name):
        parser = super(ShowShareGroupType, self).get_parser(prog_name)
        parser.add_argument(
            'share_group_type',
            metavar="<share-group-type>",
            help=_("Name or ID of the share group type to show")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_group_type = apiutils.find_resource(
            share_client.share_group_types, parsed_args.share_group_type)

        share_group_type_obj = share_client.share_group_types.get(
            share_group_type)

        formatter = parsed_args.formatter

        formatted_group_type = utils.format_share_group_type(
            share_group_type_obj, formatter)

        return (ATTRIBUTES, oscutils.get_dict_properties(
            formatted_group_type, ATTRIBUTES))


class SetShareGroupType(command.Command):
    """Set share type properties."""
    _description = _("Set share group type properties")

    log = logging.getLogger(__name__ + ".SetShareGroupType")

    def get_parser(self, prog_name):
        parser = super(SetShareGroupType, self).get_parser(prog_name)
        parser.add_argument(
            'share_group_type',
            metavar="<share-group-type>",
            help=_("Name or ID of the share group type to modify")
        )
        parser.add_argument(
            "--group-specs",
            type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
            help=_("Extra specs key and value of share group type that will be"
                   " used for share type creation. OPTIONAL: Default=None."
                   " Example: --group-specs consistent-snapshot-support=True"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        try:
            share_group_type_obj = apiutils.find_resource(
                share_client.share_group_types, parsed_args.share_group_type)
        except Exception as e:
            msg = LOG.error(_(
                "Failed to find the share group type with "
                "name or ID '%(share_group_type)s': %(e)s"),
                {'share_group_type': parsed_args.share_group_type, 'e': e})
            raise exceptions.CommandError(msg)
        kwargs = {}

        if kwargs:
            share_group_type_obj.set_keys(**kwargs)

        if parsed_args.group_specs:
            group_specs = utils.extract_group_specs(
                extra_specs={},
                specs_to_add=parsed_args.group_specs)
            try:
                share_group_type_obj.set_keys(group_specs)
            except Exception as e:
                raise exceptions.CommandError(
                    "Failed to set share group type key: %s" % e)


class UnsetShareGroupType(command.Command):
    """Unset share group type extra specs."""
    _description = _("Unset share group type extra specs")

    log = logging.getLogger(__name__ + ".UnsetShareGroupType")

    def get_parser(self, prog_name):
        parser = super(UnsetShareGroupType, self).get_parser(prog_name)
        parser.add_argument(
            'share_group_type',
            metavar="<share-group-type>",
            help=_("Name or ID of the share grouptype to modify")
        )
        parser.add_argument(
            'group_specs',
            metavar='<key>',
            nargs='+',
            help=_('Remove group specs from this share group type'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        try:
            share_group_type_obj = apiutils.find_resource(
                share_client.share_group_types, parsed_args.share_group_type)
        except Exception as e:
            msg = LOG.error(_(
                "Failed to find the share group type with "
                "name or ID '%(share_group_type)s': %(e)s"),
                {'share_group_type': parsed_args.share_group_type, 'e': e})
            raise exceptions.CommandError(msg)

        if parsed_args.group_specs:
            try:
                share_group_type_obj.unset_keys(parsed_args.group_specs)
            except Exception as e:
                raise exceptions.CommandError(
                    "Failed to remove share type group extra spec: %s" % e)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from unittest import mock
import uuid

import ddt
from osc_lib import exceptions

from manilaclient import api_versions
from manilaclient.osc.v2 import share_network_subnets as osc_share_subnets
from manilaclient.tests.unit.osc import osc_utils
from manilaclient.tests.unit.osc.v2 import fakes as manila_fakes


class TestShareNetworkSubnet(manila_fakes.TestShare):

    def setUp(self):
        super(TestShareNetworkSubnet, self).setUp()

        self.share_networks_mock = self.app.client_manager.share.share_networks
        self.share_networks_mock.reset_mock()

        self.share_subnets_mock = (
            self.app.client_manager.share.share_network_subnets)
        self.share_subnets_mock.reset_mock()


@ddt.ddt
class TestShareNetworkSubnetCreate(TestShareNetworkSubnet):

    def setUp(self):
        super(TestShareNetworkSubnetCreate, self).setUp()

        self.share_network = (
            manila_fakes.FakeShareNetwork.create_one_share_network())
        self.share_networks_mock.get.return_value = self.share_network

        self.share_network_subnet = (
            manila_fakes.FakeShareNetworkSubnet.create_one_share_subnet())
        self.share_subnets_mock.create.return_value = self.share_network_subnet

        self.cmd = osc_share_subnets.CreateShareNetworkSubnet(
            self.app, None)

        self.data = self.share_network_subnet._info.values()
        self.columns = self.share_network_subnet._info.keys()

    def test_share_network_subnet_create_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_utils.ParserException,
                          self.check_parser, self.cmd, arglist, verifylist)

    def test_share_network_subnet_create(self):
        fake_neutron_net_id = str(uuid.uuid4())
        fake_neutron_subnet_id = str(uuid.uuid4())

        arglist = [
            self.share_network.id,
            '--neutron-net-id', fake_neutron_net_id,
            '--neutron-subnet-id', fake_neutron_subnet_id,
            '--availability-zone', 'nova',
        ]
        verifylist = [
            ('share_network', self.share_network.id),
            ('neutron_net_id', fake_neutron_net_id),
            ('neutron_subnet_id', fake_neutron_subnet_id),
            ('availability_zone', 'nova'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.share_subnets_mock.create.assert_called_once_with(
            neutron_net_id=fake_neutron_net_id,
            neutron_subnet_id=fake_neutron_subnet_id,
            availability_zone='nova',
            share_network_id=self.share_network.id
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_network_subnet_create_arg_group_exception(self):
        fake_neutron_net_id = str(uuid.uuid4())

        arglist = [
            self.share_network.id,
            '--neutron-net-id', fake_neutron_net_id
        ]
        verifylist = [
            ('share_network', self.share_network.id),
            ('neutron_net_id', fake_neutron_net_id)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action,
                          parsed_args)

    @ddt.data({'check_only': False, 'restart_check': True},
              {'check_only': True, 'restart_check': True},
              {'check_only': True, 'restart_check': False})
    @ddt.unpack
    def test_share_network_subnet_create_check_api_version_exception(
            self, check_only, restart_check):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.69'
        )
        arglist = [
            self.share_network.id,
        ]
        verifylist = [
            ('share_network', self.share_network.id),
        ]
        if check_only:
            arglist.append('--check-only')
            verifylist.append(('check_only', True))
        if restart_check:
            arglist.append('--restart-check')
            verifylist.append(('restart_check', True))
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args)

    @ddt.data(True, False)
    def test_share_network_subnet_create_check(self, restart_check):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.70'
        )
        self.share_networks_mock.share_network_subnet_create_check = (
            mock.Mock(return_value=(200, {'compatible': True})))
        arglist = [
            self.share_network.id,
            '--check-only'
        ]
        verifylist = [
            ('share_network', self.share_network.id),
            ('check_only', True),
        ]
        if restart_check:
            arglist.append('--restart-check')
            verifylist.append(('restart_check', True))
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        (self.share_networks_mock.share_network_subnet_create_check
         .assert_called_once_with(
             share_network_id=self.share_network.id, neutron_net_id=None,
             neutron_subnet_id=None, availability_zone=None,
             reset_operation=restart_check))


class TestShareNetworkSubnetDelete(TestShareNetworkSubnet):

    def setUp(self):
        super(TestShareNetworkSubnetDelete, self).setUp()

        self.share_network = (
            manila_fakes.FakeShareNetwork.create_one_share_network())
        self.share_networks_mock.get.return_value = self.share_network

        self.share_network_subnets = (
            manila_fakes.FakeShareNetworkSubnet.create_share_network_subnets())

        self.cmd = osc_share_subnets.DeleteShareNetworkSubnet(
            self.app, None)

    def test_share_network_subnet_delete_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_utils.ParserException,
                          self.check_parser, self.cmd, arglist, verifylist)

    def test_share_network_subnets_delete(self):
        arglist = [
            self.share_network.id,
            self.share_network_subnets[0].id,
            self.share_network_subnets[1].id,
        ]
        verifylist = [
            ('share_network', self.share_network.id),
            ('share_network_subnet', [self.share_network_subnets[0].id,
                                      self.share_network_subnets[1].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(self.share_subnets_mock.delete.call_count,
                         len(self.share_network_subnets))
        self.assertIsNone(result)

    def test_share_network_subnet_delete_exception(self):
        arglist = [
            self.share_network.id,
            self.share_network_subnets[0].id,
        ]
        verifylist = [
            ('share_network', self.share_network.id),
            ('share_network_subnet', [self.share_network_subnets[0].id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.share_subnets_mock.delete.side_effect = exceptions.CommandError()
        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action,
                          parsed_args)


class TestShareNetworkSubnetShow(TestShareNetworkSubnet):

    def setUp(self):
        super(TestShareNetworkSubnetShow, self).setUp()

        self.share_network = (
            manila_fakes.FakeShareNetwork.create_one_share_network())
        self.share_networks_mock.get.return_value = self.share_network

        self.share_network_subnet = (
            manila_fakes.FakeShareNetworkSubnet.create_one_share_subnet())
        self.share_subnets_mock.get.return_value = self.share_network_subnet

        self.cmd = osc_share_subnets.ShowShareNetworkSubnet(
            self.app, None)

        self.data = self.share_network_subnet._info.values()
        self.columns = self.share_network_subnet._info.keys()

    def test_share_network_subnet_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_utils.ParserException,
                          self.check_parser, self.cmd, arglist, verifylist)

    def test_share_network_subnet_show(self):
        arglist = [
            self.share_network.id,
            self.share_network_subnet.id,
        ]
        verifylist = [
            ('share_network', self.share_network.id),
            ('share_network_subnet', self.share_network_subnet.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.share_subnets_mock.get.assert_called_once_with(
            self.share_network.id,
            self.share_network_subnet.id
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)



# Copyright 2015 Chuck Fouts
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

import manilaclient
from manilaclient import api_versions
from manilaclient.common import cliutils
from manilaclient import exceptions
from manilaclient.tests.unit import utils


@ddt.ddt
class APIVersionTestCase(utils.TestCase):
    def test_valid_version_strings(self):
        def _test_string(version, exp_major, exp_minor):
            v = api_versions.APIVersion(version)
            self.assertEqual(v.ver_major, exp_major)
            self.assertEqual(v.ver_minor, exp_minor)

        _test_string("1.1", 1, 1)
        _test_string("2.10", 2, 10)
        _test_string("5.234", 5, 234)
        _test_string("12.5", 12, 5)
        _test_string("2.0", 2, 0)
        _test_string("2.200", 2, 200)

    def test_null_version(self):
        v = api_versions.APIVersion()
        self.assertTrue(v.is_null())
        self.assertEqual(repr(v), "<APIVersion: null>")

    @ddt.data(
        "2",
        "200",
        "2.1.4",
        "200.23.66.3",
        "5 .3",
        "5. 3",
        "5.03",
        "02.1",
        "2.001",
        "",
        " 2.1",
        "2.1 ",
        "2.",
    )
    def test_invalid_version_strings(self, version):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, version)

    def test_version_comparisons(self):
        v1 = api_versions.APIVersion("2.0")
        v2 = api_versions.APIVersion("2.5")
        v3 = api_versions.APIVersion("5.23")
        v4 = api_versions.APIVersion("2.0")
        v5 = api_versions.APIVersion("1.0")
        v_null = api_versions.APIVersion()

        self.assertLess(v1, v2)
        self.assertGreater(v3, v2)
        self.assertNotEqual(v1, v2)
        self.assertEqual(v1, v4)
        self.assertNotEqual(v1, v_null)
        self.assertLess(v5, v1)
        self.assertLess(v5, v2)
        self.assertEqual(v_null, v_null)
        self.assertRaises(TypeError, v1.__le__, "2.1")
        self.assertRaises(TypeError, v1.__eq__, "2.1")
        self.assertRaises(TypeError, v1.__gt__, "2.1")

    def test_version_matches(self):
        v1 = api_versions.APIVersion("2.0")
        v2 = api_versions.APIVersion("2.5")
        v3 = api_versions.APIVersion("2.45")
        v4 = api_versions.APIVersion("3.3")
        v5 = api_versions.APIVersion("3.23")
        v6 = api_versions.APIVersion("2.0")
        v7 = api_versions.APIVersion("3.3")
        v8 = api_versions.APIVersion("4.0")
        v_null = api_versions.APIVersion()

        v1_25 = api_versions.APIVersion("2.5")
        v1_32 = api_versions.APIVersion("3.32")
        v1_33 = api_versions.APIVersion("3.33")
        self.assertTrue(v2.matches(v1, v3))
        self.assertTrue(v2.matches(v1, v_null))

        self.assertTrue(v1_32.matches(v1_25, v1_33))

        self.assertTrue(v1.matches(v6, v2))
        self.assertTrue(v4.matches(v2, v7))
        self.assertTrue(v4.matches(v_null, v7))
        self.assertTrue(v4.matches(v_null, v8))
        self.assertFalse(v1.matches(v2, v3))
        self.assertFalse(v5.matches(v2, v4))
        self.assertFalse(v2.matches(v3, v1))

        self.assertRaises(ValueError, v_null.matches, v1, v3)

    def test_get_string(self):
        v1_string = "3.23"
        v1 = api_versions.APIVersion(v1_string)
        self.assertEqual(v1_string, v1.get_string())

        self.assertRaises(ValueError,
                          api_versions.APIVersion().get_string)

    @ddt.data("2.0",
              "2.5",
              "2.45",
              "3.3",
              "3.23",
              "2.0",
              "3.3",
              "4.0")
    def test_representation(self, version):
        version_major, version_minor = version.split('.')
        api_version = api_versions.APIVersion(version)
        self.assertEqual(str(api_version),
                         ("API Version Major: %s, Minor: %s" %
                          (version_major, version_minor)))
        self.assertEqual(repr(api_version), "<APIVersion: %s>" % version)

    def test_is_latest(self):
        v1 = api_versions.APIVersion("1.0")
        self.assertFalse(v1.is_latest())
        v_latest = api_versions.APIVersion(api_versions.MAX_VERSION)
        self.assertTrue(v_latest.is_latest())


class GetAPIVersionTestCase(utils.TestCase):

    def test_wrong_format(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.get_api_version, "something_wrong")

    def test_wrong_major_version(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.get_api_version, "1")

    @mock.patch("manilaclient.api_versions.APIVersion")
    def test_major_and_minor_parts_is_presented(self, mock_apiversion):
        version = "2.7"
        self.assertEqual(mock_apiversion.return_value,
                         api_versions.get_api_version(version))
        mock_apiversion.assert_called_once_with(version)


class WrapsTestCase(utils.TestCase):

    def _get_obj_with_vers(self, vers):
        return mock.MagicMock(api_version=api_versions.APIVersion(vers))

    def _side_effect_of_vers_method(self, *args, **kwargs):
        m = mock.MagicMock(start_version=args[1], end_version=args[2])
        m.name = args[0]
        return m

    @mock.patch("manilaclient.utils.get_function_name")
    @mock.patch("manilaclient.api_versions.VersionedMethod")
    def test_end_version_is_none(self, mock_versioned_method, mock_name):
        func_name = 'foo'
        mock_name.return_value = func_name
        mock_versioned_method.side_effect = self._side_effect_of_vers_method

        @api_versions.wraps('2.2')
        def foo(*args, **kwargs):
            pass

        foo(self._get_obj_with_vers('2.4'))

        mock_versioned_method.assert_called_once_with(
            func_name, api_versions.APIVersion('2.2'),
            api_versions.APIVersion(api_versions.MAX_VERSION), mock.ANY)

    @mock.patch("manilaclient.utils.get_function_name")
    @mock.patch("manilaclient.api_versions.VersionedMethod")
    def test_start_and_end_version_are_presented(self, mock_versioned_method,
                                                 mock_name):
        func_name = "foo"
        mock_name.return_value = func_name
        mock_versioned_method.side_effect = self._side_effect_of_vers_method

        @api_versions.wraps("2.2", "2.6")
        def foo(*args, **kwargs):
            pass

        foo(self._get_obj_with_vers("2.4"))

        mock_versioned_method.assert_called_once_with(
            func_name, api_versions.APIVersion("2.2"),
            api_versions.APIVersion("2.6"), mock.ANY)

    @mock.patch("manilaclient.utils.get_function_name")
    @mock.patch("manilaclient.api_versions.VersionedMethod")
    def test_api_version_doesnt_match(self, mock_versioned_method, mock_name):
        func_name = "foo"
        mock_name.return_value = func_name
        mock_versioned_method.side_effect = self._side_effect_of_vers_method

        @api_versions.wraps("2.2", "2.6")
        def foo(*args, **kwargs):
            pass

        self.assertRaises(exceptions.UnsupportedVersion,
                          foo, self._get_obj_with_vers("2.1"))

        mock_versioned_method.assert_called_once_with(
            func_name, api_versions.APIVersion("2.2"),
            api_versions.APIVersion("2.6"), mock.ANY)

    def test_define_method_is_actually_called(self):
        checker = mock.MagicMock()

        @api_versions.wraps("2.2", "2.6")
        def some_func(*args, **kwargs):
            checker(*args, **kwargs)

        obj = self._get_obj_with_vers("2.4")
        some_args = ("arg_1", "arg_2")
        some_kwargs = {"key1": "value1", "key2": "value2"}

        some_func(obj, *some_args, **some_kwargs)

        checker.assert_called_once_with(*((obj,) + some_args), **some_kwargs)

    def test_cli_args_are_copied(self):

        @api_versions.wraps("2.2", "2.6")
        @cliutils.arg("name_1", help="Name of the something")
        @cliutils.arg("action_1", help="Some action")
        def some_func_1(cs, args):
            pass

        @cliutils.arg("name_2", help="Name of the something")
        @cliutils.arg("action_2", help="Some action")
        @api_versions.wraps("2.2", "2.6")
        def some_func_2(cs, args):
            pass

        args_1 = [(('name_1',), {'help': 'Name of the something'}),
                  (('action_1',), {'help': 'Some action'})]
        self.assertEqual(args_1, some_func_1.arguments)

        args_2 = [(('name_2',), {'help': 'Name of the something'}),
                  (('action_2',), {'help': 'Some action'})]
        self.assertEqual(args_2, some_func_2.arguments)


class DiscoverVersionTestCase(utils.TestCase):
    def setUp(self):
        super(DiscoverVersionTestCase, self).setUp()
        self.orig_max = manilaclient.API_MAX_VERSION
        self.orig_min = manilaclient.API_MIN_VERSION
        self.addCleanup(self._clear_fake_version)
        self.fake_client = mock.MagicMock()

    def _clear_fake_version(self):
        manilaclient.API_MAX_VERSION = self.orig_max
        manilaclient.API_MIN_VERSION = self.orig_min

    def _mock_returned_server_version(self, server_version,
                                      server_min_version):
        version_mock = mock.MagicMock(version=server_version,
                                      min_version=server_min_version,
                                      status='CURRENT')
        val = [version_mock]
        self.fake_client.services.server_api_version.return_value = val

    def test_server_is_too_new(self):
        self._mock_returned_server_version('2.7', '2.4')
        manilaclient.API_MAX_VERSION = api_versions.APIVersion("2.3")
        manilaclient.API_MIN_VERSION = api_versions.APIVersion("2.1")

        self.assertRaisesRegex(exceptions.UnsupportedVersion,
                               ".*range is '2.4' to '2.7'.*",
                               api_versions.discover_version,
                               self.fake_client,
                               api_versions.APIVersion("2.3"))
        self.assertTrue(self.fake_client.services.server_api_version.called)

    def test_server_is_too_old(self):
        self._mock_returned_server_version('2.2', '2.0')
        manilaclient.API_MAX_VERSION = api_versions.APIVersion("2.10")
        manilaclient.API_MIN_VERSION = api_versions.APIVersion("2.9")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.discover_version,
                          self.fake_client,
                          api_versions.APIVersion("2.10"))
        self.assertTrue(self.fake_client.services.server_api_version.called)

    def test_requested_version_is_less_than_server_max(self):
        self._mock_returned_server_version('2.17', '2.14')
        max_version = api_versions.APIVersion('2.15')
        manilaclient.API_MAX_VERSION = max_version
        manilaclient.API_MIN_VERSION = api_versions.APIVersion('2.12')
        version = api_versions.discover_version(self.fake_client, max_version)

        self.assertEqual(api_versions.APIVersion('2.15'), version)

    def test_requested_version_is_downgraded(self):
        server_end_version = '2.7'
        self._mock_returned_server_version(server_end_version, '2.0')
        max_version = api_versions.APIVersion("2.8")
        manilaclient.API_MAX_VERSION = max_version
        manilaclient.API_MIN_VERSION = api_versions.APIVersion("2.5")
        version = api_versions.discover_version(self.fake_client, max_version)

        self.assertEqual(api_versions.APIVersion(server_end_version), version)

    def test_server_and_client_max_are_same(self):
        self._mock_returned_server_version('2.5', '2.0')
        manilaclient.API_MAX_VERSION = api_versions.APIVersion("2.5")
        manilaclient.API_MIN_VERSION = api_versions.APIVersion("2.5")

        discovered_version = api_versions.discover_version(
            self.fake_client,
            manilaclient.API_MAX_VERSION)
        self.assertEqual("2.5", discovered_version.get_string())
        self.assertTrue(self.fake_client.services.server_api_version.called)

    def test_pre_microversion_server(self):
        self.fake_client.services.server_api_version.return_value = []
        manilaclient.API_MAX_VERSION = api_versions.APIVersion("2.5")
        manilaclient.API_MIN_VERSION = api_versions.APIVersion("2.5")
        discovered_version = api_versions.discover_version(
            self.fake_client,
            manilaclient.API_MAX_VERSION)
        self.assertEqual("1.0", discovered_version.get_string())
        self.assertTrue(self.fake_client.services.server_api_version.called)

    def test_requested_version_in_range(self):
        self._mock_returned_server_version('2.7', '2.4')
        manilaclient.API_MAX_VERSION = api_versions.APIVersion("2.11")
        manilaclient.API_MIN_VERSION = api_versions.APIVersion("2.1")

        discovered_version = api_versions.discover_version(
            self.fake_client,
            api_versions.APIVersion('2.7'))
        self.assertEqual('2.7', discovered_version.get_string())
        self.assertTrue(self.fake_client.services.server_api_version.called)

    def test_server_without_microversion(self):
        self._mock_returned_server_version(None, None)
        manilaclient.API_MAX_VERSION = api_versions.APIVersion("2.11")
        manilaclient.API_MIN_VERSION = api_versions.APIVersion("2.1")

        discovered_version = api_versions.discover_version(
            self.fake_client,
            api_versions.APIVersion('2.7'))
        self.assertEqual(api_versions.DEPRECATED_VERSION,
                         discovered_version.get_string())

        self.assertTrue(self.fake_client.services.server_api_version.called)

    def test_requested_version_is_too_old(self):
        self._mock_returned_server_version('2.5', '2.0')
        manilaclient.API_MAX_VERSION = api_versions.APIVersion("2.5")
        manilaclient.API_MIN_VERSION = api_versions.APIVersion("2.5")

        self.assertRaisesRegex(exceptions.UnsupportedVersion,
                               ".*range is '2.0' to '2.5'.*",
                               api_versions.discover_version,
                               self.fake_client,
                               api_versions.APIVersion("1.0"))



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from random import randint

import ddt
from tempest.lib.cli import output_parser
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from manilaclient import api_versions
from manilaclient.tests.functional import base
from manilaclient.tests.functional import utils

REPLICA_QUOTAS_MICROVERSION = '2.53'


def _get_share_type_quota_values(project_quota_value):
    project_quota_value = int(project_quota_value)
    if project_quota_value == -1:
        return randint(1, 999)
    elif project_quota_value == 0:
        return 0
    else:
        return project_quota_value - 1


@ddt.ddt
@utils.skip_if_microversion_not_supported("2.39")
class QuotasReadWriteTest(base.BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.microversion = "2.39"
        self.project_id = self.admin_client.get_project_id(
            self.admin_client.tenant_name)

        # Create share type
        self.share_type = self.create_share_type(
            name=data_utils.rand_name("manilaclient_functional_test"),
            driver_handles_share_servers=False,
            is_public=True,
            microversion=self.microversion
        )
        self.st_id = self.share_type["ID"]

    def _verify_current_st_quotas_equal_to(self, quotas, microversion):
        # Read share type quotas
        cmd = 'quota-show --tenant-id %s --share-type %s' % (
            self.project_id, self.st_id)
        st_quotas_raw = self.admin_client.manila(
            cmd, microversion=microversion)
        st_quotas = output_parser.details(st_quotas_raw)

        # Verify that quotas
        self.assertGreater(len(st_quotas), 3)
        for key, value in st_quotas.items():
            if key not in ('shares', 'gigabytes', 'snapshots',
                           'snapshot_gigabytes'):
                continue
            self.assertIn(key, quotas)
            self.assertEqual(int(quotas[key]), int(value))

    def _verify_current_quotas_equal_to(self, quotas, microversion):
        # Read quotas
        cmd = 'quota-show --tenant-id %s' % self.project_id
        quotas_raw = self.admin_client.manila(
            cmd, microversion=microversion)
        quotas = output_parser.details(quotas_raw)

        # Verify that quotas
        self.assertGreater(len(quotas), 3)
        for key, value in quotas.items():
            if key not in ('shares', 'gigabytes', 'snapshots',
                           'snapshot_gigabytes',
                           'share_groups', 'share_group_snapshots'):
                continue
            self.assertIn(key, quotas)
            self.assertEqual(int(quotas[key]), int(value))

    @ddt.data(*set([
        "2.40", api_versions.MAX_VERSION,
    ]))
    def test_update_quotas_for_share_groups(self, microversion):
        if not utils.is_microversion_supported(microversion):
            msg = "Microversion '%s' not supported." % microversion
            raise self.skipException(msg)

        # Get default quotas
        cmd = 'quota-defaults'
        quotas_raw = self.admin_client.manila(cmd, microversion=microversion)
        default_quotas = output_parser.details(quotas_raw)

        # Get project quotas
        cmd = 'quota-show --tenant-id %s ' % self.project_id
        quotas_raw = self.admin_client.manila(cmd, microversion=microversion)
        p_quotas = output_parser.details(quotas_raw)

        # Define custom share group quotas for project
        p_custom_quotas = {
            'share_groups': -1 if int(p_quotas['share_groups']) != -1 else 999,
            'share_group_snapshots': -1 if int(
                p_quotas['share_group_snapshots']) != -1 else 999,
        }

        # Update share group quotas for project
        cmd = ('quota-update %s --share-groups %s '
               '--share-group-snapshots %s') % (
            self.project_id,
            p_custom_quotas['share_groups'],
            p_custom_quotas['share_group_snapshots'],
        )
        self.admin_client.manila(cmd, microversion=microversion)

        # Verify quotas
        self._verify_current_quotas_equal_to(p_custom_quotas, microversion)

        # Reset quotas
        cmd = 'quota-delete --tenant-id %s --share-type %s' % (
            self.project_id, self.st_id)
        self.admin_client.manila(cmd, microversion=microversion)

        # Verify quotas after reset
        self._verify_current_quotas_equal_to(default_quotas, microversion)

        # Return project quotas back
        cmd = ('quota-update %s --share-groups %s '
               '--share-group-snapshots %s') % (
            self.project_id,
            p_quotas['share_groups'], p_quotas['share_group_snapshots'])
        self.admin_client.manila(cmd, microversion=microversion)

        # Verify quotas after reset
        self._verify_current_quotas_equal_to(p_quotas, microversion)

    @ddt.data('--share-groups', '--share-group-snapshots')
    @utils.skip_if_microversion_not_supported("2.39")
    def test_update_quotas_for_share_groups_using_too_old_microversion(self,
                                                                       arg):
        cmd = 'quota-update %s %s 13' % (self.project_id, arg)
        self.assertRaises(
            exceptions.CommandFailed,
            self.admin_client.manila,
            cmd, microversion='2.39')

    @ddt.data('--share-replicas', '--replica-gigabytes')
    @utils.skip_if_microversion_not_supported("2.52")
    def test_update_quotas_for_share_replicas_using_too_old_microversion(self,
                                                                         arg):
        cmd = 'quota-update %s %s 10' % (self.project_id, arg)
        self.assertRaises(
            exceptions.CommandFailed,
            self.admin_client.manila,
            cmd, microversion='2.52')

    @ddt.data('--share-groups', '--share-group-snapshots')
    @utils.skip_if_microversion_not_supported("2.40")
    def test_update_share_type_quotas_for_share_groups(self, arg):
        cmd = 'quota-update %s --share-type %s %s 13' % (
            self.project_id, self.st_id, arg)
        self.assertRaises(
            exceptions.CommandFailed,
            self.admin_client.manila,
            cmd, microversion='2.40')

    @ddt.data(*set([
        "2.39", "2.40", REPLICA_QUOTAS_MICROVERSION, api_versions.MAX_VERSION,
    ]))
    def test_update_share_type_quotas_positive(self, microversion):
        if not utils.is_microversion_supported(microversion):
            msg = "Microversion '%s' not supported." % microversion
            raise self.skipException(msg)

        # Get project quotas
        cmd = 'quota-show --tenant-id %s ' % self.project_id
        quotas_raw = self.admin_client.manila(cmd, microversion=microversion)
        p_quotas = output_parser.details(quotas_raw)

        # Define share type quotas
        st_custom_quotas = {
            'shares': _get_share_type_quota_values(p_quotas['shares']),
            'snapshots': _get_share_type_quota_values(p_quotas['snapshots']),
            'gigabytes': _get_share_type_quota_values(p_quotas['gigabytes']),
            'snapshot_gigabytes': _get_share_type_quota_values(
                p_quotas['snapshot_gigabytes']),
        }
        supports_share_replica_quotas = (
            api_versions.APIVersion(microversion) >= api_versions.APIVersion(
                REPLICA_QUOTAS_MICROVERSION))

        if supports_share_replica_quotas:
            st_custom_quotas['share_replicas'] = _get_share_type_quota_values(
                p_quotas['share_replicas']
            )
            st_custom_quotas['replica_gigabytes'] = (
                _get_share_type_quota_values(p_quotas['replica_gigabytes']))
            replica_params = (' --share-replicas %s '
                              '--replica-gigabytes %s') % (
                st_custom_quotas['share_replicas'],
                st_custom_quotas['replica_gigabytes'])

        # Update quotas for share type
        cmd = ('quota-update %s --share-type %s '
               '--shares %s --gigabytes %s --snapshots %s '
               '--snapshot-gigabytes %s') % (
                   self.project_id, self.st_id,
                   st_custom_quotas['shares'],
                   st_custom_quotas['gigabytes'],
                   st_custom_quotas['snapshots'],
                   st_custom_quotas['snapshot_gigabytes'])

        if supports_share_replica_quotas:
            cmd += replica_params
        self.admin_client.manila(cmd, microversion=microversion)

        # Verify share type quotas
        self._verify_current_st_quotas_equal_to(st_custom_quotas, microversion)

        # Reset share type quotas
        cmd = 'quota-delete --tenant-id %s --share-type %s' % (
            self.project_id, self.st_id)
        self.admin_client.manila(cmd, microversion=microversion)

        # Verify share type quotas after reset
        self._verify_current_st_quotas_equal_to(p_quotas, microversion)

    @utils.skip_if_microversion_not_supported("2.38")
    def test_read_share_type_quotas_with_too_old_microversion(self):
        cmd = 'quota-show --tenant-id %s --share-type %s' % (
            self.project_id, self.st_id)
        self.assertRaises(
            exceptions.CommandFailed,
            self.admin_client.manila,
            cmd, microversion='2.38')

    @utils.skip_if_microversion_not_supported("2.38")
    def test_update_share_type_quotas_with_too_old_microversion(self):
        cmd = 'quota-update --tenant-id %s --share-type %s --shares %s' % (
            self.project_id, self.st_id, '0')
        self.assertRaises(
            exceptions.CommandFailed,
            self.admin_client.manila,
            cmd, microversion='2.38')

    @utils.skip_if_microversion_not_supported("2.38")
    def test_delete_share_type_quotas_with_too_old_microversion(self):
        cmd = 'quota-delete --tenant-id %s --share-type %s' % (
            self.project_id, self.st_id)
        self.assertRaises(
            exceptions.CommandFailed,
            self.admin_client.manila,
            cmd, microversion='2.38')


@ddt.ddt
class ManilaClientTestQuotasReadOnly(base.BaseTestCase):

    def test_quota_class_show_by_admin(self):
        roles = self.parser.listing(
            self.clients['admin'].manila('quota-class-show', params='abc'))
        self.assertTableStruct(roles, ('Property', 'Value'))

    def test_quota_class_show_by_user(self):
        self.assertRaises(
            exceptions.CommandFailed,
            self.clients['user'].manila,
            'quota-class-show',
            params='abc')

    def _get_quotas(self, role, operation, microversion):
        roles = self.parser.listing(self.clients[role].manila(
            'quota-%s' % operation, microversion=microversion))
        self.assertTableStruct(roles, ('Property', 'Value'))

    @ddt.data('admin', 'user')
    @utils.skip_if_microversion_not_supported("1.0")
    def test_quota_defaults_api_1_0(self, role):
        self._get_quotas(role, "defaults", "1.0")

    @ddt.data('admin', 'user')
    @utils.skip_if_microversion_not_supported("2.0")
    def test_quota_defaults_api_2_0(self, role):
        self._get_quotas(role, "defaults", "2.0")

    @ddt.data('admin', 'user')
    @utils.skip_if_microversion_not_supported("2.6")
    def test_quota_defaults_api_2_6(self, role):
        self._get_quotas(role, "defaults", "2.6")

    @ddt.data('admin', 'user')
    @utils.skip_if_microversion_not_supported("2.7")
    def test_quota_defaults_api_2_7(self, role):
        self._get_quotas(role, "defaults", "2.7")

    @ddt.data('admin', 'user')
    @utils.skip_if_microversion_not_supported("1.0")
    def test_quota_show_api_1_0(self, role):
        self._get_quotas(role, "show", "1.0")

    @ddt.data('admin', 'user')
    @utils.skip_if_microversion_not_supported("2.0")
    def test_quota_show_api_2_0(self, role):
        self._get_quotas(role, "show", "2.0")

    @ddt.data('admin', 'user')
    @utils.skip_if_microversion_not_supported("2.6")
    def test_quota_show_api_2_6(self, role):
        self._get_quotas(role, "show", "2.6")

    @ddt.data('admin', 'user')
    @utils.skip_if_microversion_not_supported("2.7")
    def test_quota_show_api_2_7(self, role):
        self._get_quotas(role, "show", "2.7")

    @ddt.data('admin', 'user')
    @utils.skip_if_microversion_not_supported("2.25")
    def test_quota_show_api_2_25(self, role):
        self._get_quotas(role, "show  --detail", "2.25")



# Copyright 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import api_versions
from manilaclient import base

RESOURCES_NAME = 'share_servers'
RESOURCES_PATH = '/share-servers'
RESOURCE_PATH = RESOURCES_PATH + '/%s'
RESOURCE_NAME = 'share_server'
ACTION_PATH = RESOURCE_PATH + '/action'


class ShareServer(base.Resource):

    def __repr__(self):
        return "<ShareServer: %s>" % self.id

    def __getattr__(self, attr):
        if attr == 'share_network':
            attr = 'share_network_name'
        return super(ShareServer, self).__getattr__(attr)

    def delete(self):
        """Delete this share server."""
        self.manager.delete(self)

    def unmanage(self, force=False):
        """Unmanage this share server."""
        self.manager.unmanage(self, force)

    def reset_state(self, state):
        """Update the share server with the provided state."""
        self.manager.reset_state(self, state)

    def migration_check(self, host, writable, nondisruptive,
                        preserve_snapshots, new_share_network_id=None):
        """Check if the new host is suitable for migration."""
        return self.manager.migration_check(
            self, host, writable, nondisruptive,
            preserve_snapshots, new_share_network_id=new_share_network_id)

    def migration_start(self, host, writable, nondisruptive,
                        preserve_snapshots, new_share_network_id=None):
        """Migrate the share server to a new host."""
        self.manager.migration_start(
            self, host, writable, nondisruptive,
            preserve_snapshots, new_share_network_id=new_share_network_id)

    def migration_complete(self):
        """Complete migration of a share server."""
        return self.manager.migration_complete(self)

    def migration_cancel(self):
        """Attempts to cancel migration of a share server."""
        self.manager.migration_cancel(self)

    def migration_get_progress(self):
        """Obtain progress of migration of a share server."""
        return self.manager.migration_get_progress(self)

    def reset_task_state(self, task_state):
        """Reset the task state of a given share server."""
        self.manager.reset_task_state(self, task_state)


class ShareServerManager(base.ManagerWithFind):
    """Manage :class:`ShareServer` resources."""
    resource_class = ShareServer

    def get(self, server):
        """Get a share server.

        :param server: ID of the :class:`ShareServer` to get.
        :rtype: :class:`ShareServer`
        """
        server_id = base.getid(server)
        server = self._get("%s/%s" % (RESOURCES_PATH, server_id),
                           RESOURCE_NAME)
        # Split big dict 'backend_details' to separated strings
        # as next:
        # +---------------------+------------------------------------+
        # |       Property      |                Value               |
        # +---------------------+------------------------------------+
        # | details:instance_id |35203a78-c733-4b1f-b82c-faded312e537|
        # +---------------------+------------------------------------+
        for k, v in server._info["backend_details"].items():
            server._info["details:%s" % k] = v
        return server

    def details(self, server):
        """Get a share server details.

        :param server: ID of the :class:`ShareServer` to get details from.
        :rtype: list of :class:`ShareServerBackendDetails
        """
        server_id = base.getid(server)
        return self._get("%s/%s/details" % (RESOURCES_PATH, server_id),
                         "details")

    def delete(self, server):
        """Delete share server.

        :param server: ID of the :class:`ShareServer` to delete.
        """
        server_id = base.getid(server)
        self._delete(RESOURCE_PATH % server_id)

    def list(self, search_opts=None):
        """Get a list of share servers.

        :rtype: list of :class:`ShareServer`
        """
        query_string = self._build_query_string(search_opts)
        return self._list(RESOURCES_PATH + query_string, RESOURCES_NAME)

    @api_versions.wraps("2.49", "2.50")
    def manage(self, host, share_network_id, identifier, driver_options=None):

        driver_options = driver_options or {}
        body = {
            'host': host,
            'share_network_id': share_network_id,
            'identifier': identifier,
            'driver_options': driver_options,
        }

        resource_path = RESOURCE_PATH % 'manage'
        return self._create(resource_path, {'share_server': body},
                            'share_server')

    @api_versions.wraps("2.51")  # noqa
    def manage(self, host, share_network_id, identifier,  # noqa
               share_network_subnet_id=None, driver_options=None):

        driver_options = driver_options or {}
        body = {
            'host': host,
            'share_network_id': share_network_id,
            'identifier': identifier,
            'share_network_subnet_id': share_network_subnet_id,
            'driver_options': driver_options,
        }

        resource_path = RESOURCE_PATH % 'manage'
        return self._create(resource_path, {'share_server': body},
                            'share_server')

    @api_versions.wraps("2.49")
    def unmanage(self, share_server, force=False):
        return self._action("unmanage", share_server, {'force': force})

    @api_versions.wraps("2.49")
    def reset_state(self, share_server, state):
        """Update the provided share server with the provided state.

        :param share_server: either share_server object or text with its ID.
        :param state: text with new state to set for share.
        """
        return self._action("reset_status", share_server, {"status": state})

    def _action(self, action, share_server, info=None):
        """Perform a share server 'action'.

        :param action: text with action name.
        :param share_server: either share_server object or text with its ID.
        :param info: dict with data for specified 'action'.
        """
        body = {action: info}
        self.run_hooks('modify_body_for_action', body)
        url = ACTION_PATH % base.getid(share_server)
        return self.api.client.post(url, body=body)

    @api_versions.wraps("2.57")
    @api_versions.experimental_api
    def migration_check(self, share_server, host, writable, nondisruptive,
                        preserve_snapshots, new_share_network_id=None):
        """Check the share server migration to a new host

        :param share_server: either share_server object or text with its ID.
        :param host: Destination host where share server will be migrated.
        :param writable: Enforces migration to keep the shares writable.
        :param nondisruptive: Enforces migration to be nondisruptive.
        :param preserve_snapshots: Enforces migration to preserve snapshots.
        :param new_share_network_id: Specify the new share network id.
        """
        result = self._action(
            "migration_check", share_server, {
                "host": host,
                "preserve_snapshots": preserve_snapshots,
                "writable": writable,
                "nondisruptive": nondisruptive,
                "new_share_network_id": new_share_network_id,
            })
        return result[1]

    @api_versions.wraps("2.57")
    @api_versions.experimental_api
    def migration_start(self, share_server, host, writable,
                        nondisruptive, preserve_snapshots,
                        new_share_network_id=None):
        """Migrates share server to a new host

        :param share_server: either share_server object or text with its ID.
        :param host: Destination host where share server will be migrated.
        :param writable: Enforces migration to keep the shares writable.
        :param nondisruptive: Enforces migration to be nondisruptive.
        :param preserve_snapshots: Enforces migration to preserve snapshots.
        :param new_share_network_id: Specify the new share network id.
        """
        return self._action(
            "migration_start", share_server, {
                "host": host,
                "writable": writable,
                "nondisruptive": nondisruptive,
                "preserve_snapshots": preserve_snapshots,
                "new_share_network_id": new_share_network_id,
            })

    @api_versions.wraps("2.57")
    @api_versions.experimental_api
    def reset_task_state(self, share_server, task_state):
        """Update the provided share server with the provided task state.

        :param share_server: either share_server object or text with its ID.
        :param task_state: text with new task state to set for share.
        """
        return self._action('reset_task_state', share_server,
                            {"task_state": task_state})

    @api_versions.wraps("2.57")
    @api_versions.experimental_api
    def migration_complete(self, share_server):
        """Completes migration for a given share server.

        :param share_server: either share_server object or text with its ID.
        """
        result = self._action('migration_complete', share_server)
        # NOTE(dviroel): result[0] is response code, result[1] is dict body
        return result[1]

    @api_versions.wraps("2.57")
    @api_versions.experimental_api
    def migration_cancel(self, share_server):
        """Attempts to cancel migration for a given share server.

        :param share_server: either share_server object or text with its ID.
        """
        return self._action('migration_cancel',
                            share_server)

    @api_versions.wraps("2.57")
    @api_versions.experimental_api
    def migration_get_progress(self, share_server):
        """Obtains progress of share migration for a given share server.

        :param share_server: either share_server object or text with its ID.
        """
        result = self._action('migration_get_progress', share_server)
        # NOTE(felipefutty): result[0] is response code, result[1] is dict body
        return result[1]



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from openstackclient.identity import common as identity_common
from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as osc_utils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.common.apiclient import utils as apiutils
from manilaclient.common import constants

LOG = logging.getLogger(__name__)


class DeleteShareServer(command.Command):
    """Delete one or more share servers (Admin only)"""
    _description = _(
        "Delete one or more share servers")

    def get_parser(self, prog_name):
        parser = super(DeleteShareServer, self).get_parser(prog_name)
        parser.add_argument(
            "share_servers",
            metavar="<share-server>",
            nargs="+",
            help=_("ID(s) of the server(s) to delete")
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait for share server deletion.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for server in parsed_args.share_servers:
            try:
                server_obj = osc_utils.find_resource(
                    share_client.share_servers, server)

                share_client.share_servers.delete(server_obj)
                if parsed_args.wait:
                    if not osc_utils.wait_for_delete(
                            manager=share_client.share_servers,
                            res_id=server_obj.id):
                        result += 1

            except Exception as e:
                result += 1
                LOG.error(_(
                    "Failed to delete a share server with "
                    "ID '%(server)s': %(e)s"),
                    {'server': server, 'e': e})

        if result > 0:
            total = len(parsed_args.share_servers)
            msg = f'Failed to delete {result} servers out of {total}.'
            raise exceptions.CommandError(_(msg))


class ShowShareServer(command.ShowOne):
    """Show share server (Admin only)."""
    _description = _("Show details about a share server (Admin only).")

    def get_parser(self, prog_name):
        parser = super(ShowShareServer, self).get_parser(prog_name)
        parser.add_argument(
            "share_server",
            metavar="<share-server>",
            help=_("ID of share server.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_server = osc_utils.find_resource(
            share_client.share_servers,
            parsed_args.share_server)

        # All 'backend_details' data already present as separated strings,
        # so remove big dict from view.
        if "backend_details" in share_server._info:
            del share_server._info["backend_details"]

        share_server._info.pop('links', None)
        return self.dict2columns(share_server._info)


class ListShareServer(command.Lister):
    """List all share servers (Admin only)."""
    _description = _("List all share servers (Admin only).")

    def get_parser(self, prog_name):
        parser = super(ListShareServer, self).get_parser(prog_name)
        parser.add_argument(
            '--host',
            metavar='<hostname>',
            default=None,
            help=_('Filter results by name of host.'),
        )
        parser.add_argument(
            '--status',
            metavar="<status>",
            default=None,
            help=_('Filter results by status.')
        )
        parser.add_argument(
            '--share-network',
            metavar='<share-network>',
            default=None,
            help=_('Filter results by share network name or ID.'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            default=None,
            help=_('Filter results by project name or ID.')
        )
        parser.add_argument(
            '--share-network-subnet',
            metavar='<share-network-subnet>',
            type=str,
            default=None,
            help=_("Filter results by share network subnet that the "
                   "share server's network allocation exists within. "
                   "Available for microversion >= 2.51 (Optional, "
                   "Default=None)")
        )
        identity_common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        project_id = None
        if parsed_args.project:
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain).id

        if (parsed_args.share_network_subnet and
                share_client.api_version < api_versions.APIVersion("2.51")):
            raise exceptions.CommandError(
                "Share network subnet can be specified only with manila API "
                "version >= 2.51"
            )

        columns = [
            'ID',
            'Host',
            'Status',
            'Share Network ID',
            'Project ID',
        ]

        search_opts = {
            'status': parsed_args.status,
            'host': parsed_args.host,
            'project_id': project_id,
        }

        if parsed_args.share_network:
            share_network_id = osc_utils.find_resource(
                share_client.share_networks,
                parsed_args.share_network).id
            search_opts['share_network'] = share_network_id

        if parsed_args.share_network_subnet:
            search_opts['share_network_subnet_id'] = (
                parsed_args.share_network_subnet)

        share_servers = share_client.share_servers.list(
            search_opts=search_opts)

        data = (osc_utils.get_dict_properties(
            share_server._info, columns) for share_server in share_servers)

        return (columns, data)


class AdoptShareServer(command.ShowOne):
    """Adopt share server not handled by Manila (Admin only)."""

    _description = _("Adopt share server not handled by Manila (Admin only).")

    def get_parser(self, prog_name):
        parser = super(AdoptShareServer, self).get_parser(prog_name)
        parser.add_argument(
            'host',
            metavar='<host>',
            type=str,
            help=_('Backend name as "<node_hostname>@<backend_name>".')
        )
        parser.add_argument(
            "share_network",
            metavar="<share-network>",
            help=_("Share network where share server has network "
                   "allocations in.")
        )
        parser.add_argument(
            'identifier',
            metavar='<identifier>',
            type=str,
            help=_("A driver-specific share server identifier required "
                   "by the driver to manage the share server.")
        )
        parser.add_argument(
            '--driver-options',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            default={},
            help=_("One or more driver-specific key=value pairs that may be "
                   "necessary to manage the share server (Optional, "
                   "Default=None).")
        )
        parser.add_argument(
            '--share-network-subnet',
            type=str,
            metavar='<share-network-subnet>',
            default=None,
            help="Share network subnet where share server has network  "
                 "allocations in.The default subnet will be used if "
                 "it's not specified. Available for microversion "
                 ">= 2.51 (Optional, Default=None)."
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            help=_("Wait until share server is adopted")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_network = None
        if parsed_args.share_network:
            share_network = osc_utils.find_resource(
                share_client.share_networks,
                parsed_args.share_network).id

        share_network_subnet = None
        if (parsed_args.share_network_subnet and
                share_client.api_version < api_versions.APIVersion("2.51")):
            raise exceptions.CommandError(
                "Share network subnet can be specified only with manila API "
                "version >= 2.51"
            )
        elif parsed_args.share_network_subnet:
            share_network_subnet = share_client.share_network_subnets.get(
                share_network, parsed_args.share_network_subnet).id

        share_server = share_client.share_servers.manage(
            host=parsed_args.host,
            share_network_id=share_network,
            identifier=parsed_args.identifier,
            driver_options=parsed_args.driver_options,
            share_network_subnet_id=share_network_subnet
        )

        if parsed_args.wait:
            if not osc_utils.wait_for_status(
                status_f=share_client.share_servers.get,
                res_id=share_server.id,
                success_status=['active'],
                error_status=['manage_error', 'error']
            ):
                LOG.error(_("ERROR: Share server is in error state."))

            share_server = osc_utils.find_resource(share_client.share_servers,
                                                   share_server.id)

        share_server._info.pop('links', None)

        # All 'backend_details' data already present as separated strings,
        # so remove big dict from view.
        share_server._info.pop("backend_details", None)

        return self.dict2columns(share_server._info)


class AbandonShareServer(command.Command):
    """Remove one or more share servers (Admin only)."""

    _description = _("Remove one or more share server(s) (Admin only).")

    def get_parser(self, prog_name):
        parser = super(AbandonShareServer, self).get_parser(prog_name)
        parser.add_argument(
            "share_server",
            metavar="<share-server>",
            nargs='+',
            help=_("ID of the server(s) to be abandoned.")
        )
        parser.add_argument(
            "--force",
            action='store_true',
            default=False,
            help=_("Enforces the unmanage share server operation, even "
                   "if the backend driver does not support it.")
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait until share server is abandoned")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for server in parsed_args.share_server:
            try:
                server_obj = osc_utils.find_resource(
                    share_client.share_servers,
                    server)
                kwargs = {}
                if parsed_args.force:
                    kwargs['force'] = parsed_args.force
                share_client.share_servers.unmanage(
                    server_obj, **kwargs)

                if parsed_args.wait:
                    if not osc_utils.wait_for_delete(
                            manager=share_client.share_servers,
                            res_id=server_obj.id):
                        result += 1

            except Exception as e:
                result += 1
                LOG.error(_(
                    "Failed to abandon share server with "
                    "ID '%(server)s': %(e)s"),
                    {'server': server, 'e': e})

        if result > 0:
            total = len(parsed_args.share_server)
            msg = f'Failed to abandon {result} of {total} servers.'
            raise exceptions.CommandError(_(msg))


class SetShareServer(command.Command):
    """Set share server properties."""

    _description = _("Set share server properties (Admin only).")

    def get_parser(self, prog_name):
        parser = super(SetShareServer, self).get_parser(prog_name)
        allowed_update_choices = [
            'unmanage_starting', 'server_migrating_to', 'error',
            'unmanage_error', 'manage_error', 'inactive', 'active',
            'server_migrating', 'manage_starting', 'deleting',
            'network_change']
        allowed_update_choices_str = ', '.join(allowed_update_choices)
        parser.add_argument(
            "share_server",
            metavar="<share-server>",
            help=_("ID of the share server to modify.")
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            required=False,
            default=constants.STATUS_ACTIVE,
            help=_("Assign a status to the share server. Options "
                   "include: %s. If no state is "
                   "provided, active will be "
                   "used." % allowed_update_choices_str)
        )
        parser.add_argument(
            '--task-state',
            metavar="<task-state>",
            required=False,
            default=None,
            help=_("Indicate which task state to assign the share server. "
                   "Options include migration_starting, migration_in_progress,"
                   " migration_completing, migration_success, migration_error,"
                   " migration_cancelled, migration_driver_in_progress, "
                   "migration_driver_phase1_done, data_copying_starting, "
                   "data_copying_in_progress, data_copying_completing, "
                   "data_copying_completed, data_copying_cancelled, "
                   "data_copying_error. ")
        )
        return parser

    def take_action(self, parsed_args):
        if not parsed_args.status and not parsed_args.task_state:
            msg = (_("A status or a task state should be provided for this "
                     "command."))
            LOG.error(msg)
            raise exceptions.CommandError(msg)
        share_client = self.app.client_manager.share

        share_server = osc_utils.find_resource(
            share_client.share_servers,
            parsed_args.share_server)

        if parsed_args.status:
            try:
                share_client.share_servers.reset_state(
                    share_server,
                    parsed_args.status
                )
            except Exception as e:
                msg = (_(
                    "Failed to set status '%(status)s': %(exception)s"),
                    {'status': parsed_args.status, 'exception': e})
                LOG.error(msg)
                raise exceptions.CommandError(msg)

        if parsed_args.task_state:
            if share_client.api_version < api_versions.APIVersion("2.57"):
                raise exceptions.CommandError(
                    "Setting the state of a share server is only available "
                    "with manila API version >= 2.57")
            else:
                result = 0
                try:
                    share_client.share_servers.reset_task_state(
                        share_server, parsed_args.task_state)
                except Exception as e:
                    LOG.error(_("Failed to update share server task state "
                                "%s"), e)
                    result += 1

            if result > 0:
                raise exceptions.CommandError(_("One or more of the "
                                              "reset operations failed"))


class ShareServerMigrationCancel(command.Command):
    """Attempts to cancel migration for a given share server

    :param share_server: either share_server object or text with its ID.

    """

    _description = _("Cancels migration of a given share server when copying")

    def get_parser(self, prog_name):
        parser = super(ShareServerMigrationCancel, self).get_parser(prog_name)
        parser.add_argument(
            'share_server',
            metavar='<share_server>',
            help=_('ID of share server to cancel migration.')
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        share_server = osc_utils.find_resource(
            share_client.share_servers,
            parsed_args.share_server)
        if share_client.api_version >= api_versions.APIVersion("2.57"):
            share_server.migration_cancel()
        else:
            raise exceptions.CommandError(
                "Share Server Migration cancel is only available "
                "with manila API version >= 2.57")


class ShareServerMigrationComplete(command.Command):
    """Completes migration for a given share server (Admin only, Experimental).

    """
    _description = _("Completes migration for a given share server")

    def get_parser(self, prog_name):
        parser = super(ShareServerMigrationComplete, self).get_parser(
            prog_name)
        parser.add_argument(
            'share_server',
            metavar='<share_server>',
            help=_('ID of share server to complete migration.')
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        share_server = osc_utils.find_resource(
            share_client.share_servers,
            parsed_args.share_server)
        if share_client.api_version >= api_versions.APIVersion("2.57"):
            share_server.migration_complete()
        else:
            raise exceptions.CommandError(
                "Share Server Migration complete is only available "
                "with manila API version >= 2.57")


class ShareServerMigrationShow(command.ShowOne):
    """Obtains progress of share migration for a given share server.

    (Admin only, Experimental).

    :param share_server: either share_server object or text with its ID.

    """

    _description = _(
        "Gets migration progress of a given share server when copying")

    def get_parser(self, prog_name):
        parser = super(ShareServerMigrationShow, self).get_parser(prog_name)
        parser.add_argument(
            'share_server',
            metavar='<share_server>',
            help='ID of share server to show migration progress for.'
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        if share_client.api_version >= api_versions.APIVersion("2.57"):
            share_server = osc_utils.find_resource(
                share_client.share_servers,
                parsed_args.share_server)
            result = share_server.migration_get_progress()
            return self.dict2columns(result)
        else:
            raise exceptions.CommandError(
                "Share Server Migration show is only available "
                "with manila API version >= 2.57")


class ShareServerMigrationStart(command.ShowOne):
    """Migrates share server to a new host (Admin only, Experimental)."""

    _description = _("Migrates share server to a new host.")

    def get_parser(self, prog_name):
        parser = super(ShareServerMigrationStart, self).get_parser(prog_name)
        parser.add_argument(
            'share_server',
            metavar='<share_server>',
            help=_('ID of share server to start migration.')
        )
        parser.add_argument(
            'host',
            metavar='<host@backend>',
            help=_("Destination to migrate the share server to. Use "
                   "the format '<node_hostname>@<backend_name>'.")
        )
        parser.add_argument(
            '--preserve-snapshots',
            metavar='<True|False>',
            choices=['True', 'False'],
            required=True,
            help=_("Set to True if snapshots must be preserved at "
                   "the migration destination.")
        )
        parser.add_argument(
            '--writable',
            metavar='<True|False>',
            choices=['True', 'False'],
            required=True,
            help=_("Enforces migration to keep all its shares writable "
                   "while contents are being moved.")
        )
        parser.add_argument(
            '--nondisruptive',
            metavar='<True|False>',
            choices=['True', 'False'],
            required=True,
            help=_("Enforces migration to be nondisruptive.")
        )
        parser.add_argument(
            '--new-share-network',
            metavar='<new_share_network>',
            required=False,
            default=None,
            help=_('Specify a new share network for the share server. Do not '
                   'specify this parameter if the migrating share server has '
                   'to be retained within its current share network.',)
        )
        parser.add_argument(
            '--check-only',
            action='store_true',
            default=False,
            help=_("Run a dry-run of the share server migration. ")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        share_server = osc_utils.find_resource(
            share_client.share_servers,
            parsed_args.share_server)

        if share_client.api_version >= api_versions.APIVersion("2.57"):
            new_share_net_id = None
            result = None
            if parsed_args.new_share_network:
                new_share_net_id = apiutils.find_resource(
                    share_client.share_networks,
                    parsed_args.new_share_network).id
            if parsed_args.check_only:
                result = share_server.migration_check(
                    parsed_args.host, parsed_args.writable,
                    parsed_args.nondisruptive, parsed_args.preserve_snapshots,
                    new_share_net_id
                )
            if result:
                return self.dict2columns(result)
            else:
                share_server.migration_start(parsed_args.host,
                                             parsed_args.writable,
                                             parsed_args.nondisruptive,
                                             parsed_args.preserve_snapshots,
                                             new_share_net_id)
                return ({}, {})
        else:
            raise exceptions.CommandError(
                "Share Server Migration is only available "
                "with manila API version >= 2.57")



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.common._i18n import _

LOG = logging.getLogger(__name__)


class CreateShareSecurityService(command.ShowOne):
    """Create security service used by project."""
    _description = _("Create security service used by project.")

    def get_parser(self, prog_name):
        parser = super(CreateShareSecurityService, self).get_parser(prog_name)
        parser.add_argument(
            'type',
            metavar='<type>',
            default=None,
            choices=['ldap', 'kerberos', 'active_directory'],
            help=_("Security service type. Possible options are: "
                   "'ldap', 'kerberos', 'active_directory'.")
        )
        parser.add_argument(
            '--dns-ip',
            metavar='<dns-ip>',
            default=None,
            help=_("DNS IP address of the security service used "
                   "inside project's network.")
        )
        parser.add_argument(
            '--ou',
            metavar='<ou>',
            default=None,
            help=_("Security service OU (Organizational Unit). "
                   "Available only for microversion >= 2.44.")
        )
        parser.add_argument(
            '--server',
            metavar='<server>',
            default=None,
            help=_("Security service IP address or hostname.")
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            default=None,
            help=_("Security service domain.")
        )
        parser.add_argument(
            '--user',
            metavar='<user',
            default=None,
            help=_("Security service user or group used by project.")
        )
        parser.add_argument(
            '--password',
            metavar='<password>',
            default=None,
            help=_("Password used by user.")
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help=_("Security service name.")
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            default=None,
            help=_("Security service description.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        kwargs = {
            'dns_ip': parsed_args.dns_ip,
            'server': parsed_args.server,
            'domain': parsed_args.domain,
            'user': parsed_args.user,
            'password': parsed_args.password,
            'name': parsed_args.name,
            'description': parsed_args.description,
        }

        if share_client.api_version >= api_versions.APIVersion("2.44"):
            kwargs['ou'] = parsed_args.ou

        elif parsed_args.ou:
            raise exceptions.CommandError(
                "Defining a security service Organizational Unit is "
                "available only for microversion >= 2.44")

        security_service = share_client.security_services.create(
            parsed_args.type, **kwargs)

        return self.dict2columns(security_service._info)


class DeleteShareSecurityService(command.Command):
    """Delete one or more security services."""
    _description = _("Delete one or more security services.")

    def get_parser(self, prog_name):
        parser = super(DeleteShareSecurityService, self).get_parser(prog_name)
        parser.add_argument(
            'security_service',
            metavar='<security-service>',
            nargs="+",
            help=_("Name or ID of the security service(s) to delete.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for security_service in parsed_args.security_service:
            try:
                security_service_obj = oscutils.find_resource(
                    share_client.security_services,
                    security_service)
                share_client.security_services.delete(
                    security_service_obj)

            except Exception as e:
                result += 1
                LOG.error(f"Failed to delete security service with "
                          f"name or ID {security_service}: {e}")

        if result > 0:
            total = len(parsed_args.security_service)
            msg = (f"{result} of {total} security services failed "
                   f"to be deleted.")
            raise exceptions.CommandError(msg)


class ShowShareSecurityService(command.ShowOne):
    """Show security service."""
    _description = _("Show security service.")

    def get_parser(self, prog_name):
        parser = super(ShowShareSecurityService, self).get_parser(prog_name)
        parser.add_argument(
            'security_service',
            metavar='<security-service>',
            help=_("Security service name or ID to show.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        security_service = oscutils.find_resource(
            share_client.security_services,
            parsed_args.security_service)

        data = security_service._info
        if parsed_args.formatter == 'table':
            if 'share_networks' in data.keys():
                data['share_networks'] = "\n".join(
                    data['share_networks'])

        return self.dict2columns(data)


class SetShareSecurityService(command.Command):
    """Set security service."""
    _description = _("Set security service.")

    def get_parser(self, prog_name):
        parser = super(SetShareSecurityService, self).get_parser(prog_name)
        parser.add_argument(
            'security_service',
            metavar='<security-service>',
            help=_("Security service name or ID.")
        )
        parser.add_argument(
            '--dns-ip',
            metavar='<dns-ip>',
            default=None,
            help=_("Set DNS IP address used inside project's network.")
        )
        parser.add_argument(
            '--ou',
            metavar='<ou>',
            default=None,
            help=_("Set security service OU (Organizational Unit). "
                   "Available only for microversion >= 2.44.")
        )
        parser.add_argument(
            '--server',
            metavar='<server>',
            default=None,
            help=_("Set security service IP address or hostname.")
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            default=None,
            help=_("Set security service domain.")
        )
        parser.add_argument(
            '--user',
            metavar='<user',
            default=None,
            help=_("Set security service user or group used by project.")
        )
        parser.add_argument(
            '--password',
            metavar='<password>',
            default=None,
            help=_("Set password used by user.")
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help=_("Set security service name.")
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            default=None,
            help=_("Set security service description.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        security_service = oscutils.find_resource(
            share_client.security_services,
            parsed_args.security_service)

        kwargs = {
            'dns_ip': parsed_args.dns_ip,
            'server': parsed_args.server,
            'domain': parsed_args.domain,
            'user': parsed_args.user,
            'password': parsed_args.password,
            'name': parsed_args.name,
            'description': parsed_args.description,
        }

        if share_client.api_version >= api_versions.APIVersion("2.44"):
            kwargs['ou'] = parsed_args.ou

        elif parsed_args.ou:
            raise exceptions.CommandError(_(
                "Setting a security service Organizational Unit is "
                "available only for microversion >= 2.44"))
        try:
            security_service.update(**kwargs)
        except Exception as e:
            raise exceptions.CommandError(
                f"One or more set operations failed: {e}")


class UnsetShareSecurityService(command.Command):
    """Unset security service."""
    _description = _("Unset security service.")

    def get_parser(self, prog_name):
        parser = super(UnsetShareSecurityService, self).get_parser(prog_name)
        parser.add_argument(
            'security_service',
            metavar='<security-service>',
            help=_("Security service name or ID.")
        )
        parser.add_argument(
            '--dns-ip',
            action='store_true',
            help=_("Unset DNS IP address used inside project's network.")
        )
        parser.add_argument(
            '--ou',
            action='store_true',
            help=_("Unset security service OU (Organizational Unit). "
                   "Available only for microversion >= 2.44.")
        )
        parser.add_argument(
            '--server',
            action='store_true',
            help=_("Unset security service IP address or hostname.")
        )
        parser.add_argument(
            '--domain',
            action='store_true',
            help=_("Unset security service domain.")
        )
        parser.add_argument(
            '--user',
            action='store_true',
            help=_("Unset security service user or group used by project.")
        )
        parser.add_argument(
            '--password',
            action='store_true',
            help=_("Unset password used by user.")
        )
        parser.add_argument(
            '--name',
            action='store_true',
            help=_("Unset security service name.")
        )
        parser.add_argument(
            '--description',
            action='store_true',
            help=_("Unset security service description.")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        security_service = oscutils.find_resource(
            share_client.security_services,
            parsed_args.security_service)

        kwargs = {}
        args = ['dns_ip', 'server', 'domain', 'user', 'password',
                'name', 'description']
        for arg in args:
            if getattr(parsed_args, arg):
                # the SDK unsets a value if it is an empty string
                kwargs[arg] = ''

        if (parsed_args.ou and
                share_client.api_version >= api_versions.APIVersion("2.44")):
            # the SDK unsets a value if it is an empty string
            kwargs['ou'] = ''

        elif parsed_args.ou:
            raise exceptions.CommandError(_(
                "Unsetting a security service Organizational Unit is "
                "available only for microversion >= 2.44"))
        try:
            security_service.update(**kwargs)
        except Exception as e:
            raise exceptions.CommandError(
                f"One or more unset operations failed: {e}")


class ListShareSecurityService(command.Lister):
    """List security services."""
    _description = _("List security services.")

    def get_parser(self, prog_name):
        parser = super(ListShareSecurityService, self).get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            help=_("Display information from all projects (Admin only).")
        )
        parser.add_argument(
            '--share-network',
            metavar='<share-network>',
            default=None,
            help=_("Filter results by share network name or ID.")
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            default=None,
            help=_("Filter results by status.")
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help=_("Filter results by security service name.")
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            default=None,
            help=_("Filter results by security service type.")
        )
        parser.add_argument(
            '--user',
            metavar='<user',
            default=None,
            help=_("Filter results by security service user or group "
                   "used by project.")
        )
        parser.add_argument(
            '--dns-ip',
            metavar='<dns-ip>',
            default=None,
            help=_("Filter results by DNS IP address used inside "
                   "project's network.")
        )
        parser.add_argument(
            '--ou',
            metavar='<ou>',
            default=None,
            help=_("Filter results by security service OU "
                   "(Organizational Unit). "
                   "Available only for microversion >= 2.44.")
        )
        parser.add_argument(
            '--server',
            metavar='<server>',
            default=None,
            help=_("Filter results by security service IP "
                   "address or hostname.")
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            default=None,
            help=_("Filter results by security service domain.")
        )
        parser.add_argument(
            '--detail',
            action='store_true',
            help=_("Show detailed information about filtered "
                   "security services.")
        )
        parser.add_argument(
            "--limit",
            metavar="<num-security-services>",
            type=int,
            default=None,
            action=parseractions.NonNegativeAction,
            help=_("Limit the number of security services returned")
        )
        parser.add_argument(
            "--marker",
            metavar="<security-service>",
            help=_("The last security service ID of the previous page")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        columns = ['ID', 'Name', 'Status', 'Type']

        if parsed_args.all_projects:
            columns.append('Project ID')

        if parsed_args.detail:
            columns.append('Share Networks')

        search_opts = {
            'all_tenants': parsed_args.all_projects,
            'status': parsed_args.status,
            'name': parsed_args.name,
            'type': parsed_args.type,
            'user': parsed_args.user,
            'dns_ip': parsed_args.dns_ip,
            'server': parsed_args.server,
            'domain': parsed_args.domain,
            'offset': parsed_args.marker,
            'limit': parsed_args.limit,
        }

        if (parsed_args.ou and
                share_client.api_version >= api_versions.APIVersion("2.44")):
            search_opts['ou'] = parsed_args.ou

        elif parsed_args.ou:
            raise exceptions.CommandError(_(
                "Filtering results by security service Organizational Unit is "
                "available only for microversion >= 2.44"))

        if parsed_args.share_network:
            search_opts['share_network_id'] = oscutils.find_resource(
                share_client.share_networks,
                parsed_args.share_network).id

        data = share_client.security_services.list(
            search_opts=search_opts,
            detailed=parsed_args.detail
        )

        return (
            columns,
            (oscutils.get_item_properties(s, columns) for s in data)
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from debtcollector import removals

from keystoneauth1 import adapter
from keystoneauth1 import session
from keystoneclient import client as ks_client

import manilaclient
from manilaclient.common import constants
from manilaclient.common import httpclient
from manilaclient import exceptions
from manilaclient.v2 import limits
from manilaclient.v2 import quota_classes
from manilaclient.v2 import quotas
from manilaclient.v2 import scheduler_stats
from manilaclient.v2 import security_services
from manilaclient.v2 import services
from manilaclient.v2 import share_networks
from manilaclient.v2 import share_servers
from manilaclient.v2 import share_snapshots
from manilaclient.v2 import share_type_access
from manilaclient.v2 import share_types
from manilaclient.v2 import shares


@removals.removed_class("Client", message="Please use 'v2.Client' instead",
                        removal_version='2.0.0')
class Client(object):
    """Top-level object to access the OpenStack Manila API.

    Create an instance with your creds::

        >>> client = Client(username=USERNAME,
                            password=PASSWORD,
                            project_name=PROJECT_NAME,
                            auth_url=AUTH_URL)

    Or, alternatively, you can create a client instance using the
    keystoneauth1.session API::

        >>> from keystoneclient.auth.identity import v2
        >>> from keystoneauth1 import session
        >>> from manilaclient import client
        >>> auth = v2.Password(auth_url=AUTH_URL,
                               username=USERNAME,
                               password=PASSWORD,
                               project_name=PROJECT_ID)
        >>> sess = session.Session(auth=auth)
        >>> manila = client.Client(VERSION, session=sess)

    Then call methods on its managers::

        >>> client.shares.list()
        ...
    """
    def __init__(self, username=None, project_id=None, auth_url=None,
                 insecure=False, timeout=None, tenant_id=None,
                 project_name=None, region_name=None,
                 endpoint_type='publicURL', extensions=None,
                 service_type=constants.V1_SERVICE_TYPE, service_name=None,
                 retries=None, http_log_debug=False, input_auth_token=None,
                 session=None, auth=None, cacert=None,
                 service_catalog_url=None, user_agent='python-manilaclient',
                 use_keyring=False, force_new_token=False,
                 cached_token_lifetime=300,
                 api_version=manilaclient.API_DEPRECATED_VERSION,
                 user_id=None,
                 user_domain_id=None,
                 user_domain_name=None,
                 project_domain_id=None,
                 project_domain_name=None,
                 cert=None,
                 password=None,
                 **kwargs):

        self.username = username
        self.password = password
        self.tenant_id = tenant_id or project_id
        self.tenant_name = project_name

        self.user_id = user_id
        self.project_id = project_id or tenant_id
        self.project_name = project_name
        self.user_domain_id = user_domain_id
        self.user_domain_name = user_domain_name
        self.project_domain_id = project_domain_id
        self.project_domain_name = project_domain_name

        self.endpoint_type = endpoint_type
        self.auth_url = auth_url
        self.region_name = region_name

        self.cacert = cacert
        self.cert = cert
        self.insecure = insecure

        self.use_keyring = use_keyring
        self.force_new_token = force_new_token
        self.cached_token_lifetime = cached_token_lifetime

        if input_auth_token and not service_catalog_url:
            msg = ("For token-based authentication you should "
                   "provide 'input_auth_token' and 'service_catalog_url'.")
            raise exceptions.ClientException(msg)

        self.project_id = tenant_id if tenant_id is not None else project_id
        self.keystone_client = None
        self.session = session

        # NOTE(u_glide): token authorization has highest priority.
        # That's why session and/or password will be ignored
        # if token is provided.
        if not input_auth_token:
            if session:
                self.keystone_client = adapter.LegacyJsonAdapter(
                    session=session,
                    auth=auth,
                    interface=endpoint_type,
                    service_type=service_type,
                    service_name=service_name,
                    region_name=region_name)
                input_auth_token = self.keystone_client.session.get_token(auth)

            else:
                self.keystone_client = self._get_keystone_client()
                input_auth_token = self.keystone_client.auth_token

        if not input_auth_token:
            raise RuntimeError("Not Authorized")

        if session and not service_catalog_url:
            service_catalog_url = self.keystone_client.session.get_endpoint(
                auth, interface=endpoint_type,
                service_type=service_type)
        elif not service_catalog_url:
            catalog = self.keystone_client.service_catalog.get_endpoints(
                service_type)
            for catalog_entry in catalog.get(service_type, []):
                if (catalog_entry.get("interface") == (
                        endpoint_type.lower().split("url")[0]) or
                        catalog_entry.get(endpoint_type)):
                    if (region_name and not region_name == (
                            catalog_entry.get(
                                "region",
                                catalog_entry.get("region_id")))):
                        continue
                    service_catalog_url = catalog_entry.get(
                        "url", catalog_entry.get(endpoint_type))
                    break

        if not service_catalog_url:
            raise RuntimeError("Could not find Manila endpoint in catalog")

        self.api_version = api_version
        self.client = httpclient.HTTPClient(service_catalog_url,
                                            input_auth_token,
                                            user_agent,
                                            insecure=insecure,
                                            cacert=cacert,
                                            cert=cert,
                                            timeout=timeout,
                                            retries=retries,
                                            http_log_debug=http_log_debug,
                                            api_version=self.api_version)

        self.limits = limits.LimitsManager(self)
        self.services = services.ServiceManager(self)
        self.security_services = security_services.SecurityServiceManager(self)
        self.share_networks = share_networks.ShareNetworkManager(self)

        self.quota_classes = quota_classes.QuotaClassSetManager(self)
        self.quotas = quotas.QuotaSetManager(self)

        self.shares = shares.ShareManager(self)
        self.share_snapshots = share_snapshots.ShareSnapshotManager(self)

        self.share_types = share_types.ShareTypeManager(self)
        self.share_type_access = share_type_access.ShareTypeAccessManager(self)
        self.share_servers = share_servers.ShareServerManager(self)
        self.pools = scheduler_stats.PoolManager(self)

        self._load_extensions(extensions)

    def _load_extensions(self, extensions):
        if not extensions:
            return

        for extension in extensions:
            if extension.manager_class:
                setattr(self, extension.name, extension.manager_class(self))

    def _get_keystone_client(self):
        # First create a Keystone session
        if self.insecure:
            verify = False
        else:
            verify = self.cacert or True
        ks_session = session.Session(verify=verify, cert=self.cert)

        # Discover the supported keystone versions using the given url
        ks_discover = session.discover.Discover(ks_session, self.auth_url)

        # Inspect the auth_url to see the supported version. If both v3 and v2
        # are supported, then use the highest version if possible.
        v2_auth_url = ks_discover.url_for('v2.0')
        v3_auth_url = ks_discover.url_for('v3.0')

        if v3_auth_url:
            keystone_client = ks_client.Client(
                session=ks_session,
                version=(3, 0),
                auth_url=v3_auth_url,
                username=self.username,
                password=self.password,
                user_id=self.user_id,
                user_domain_name=self.user_domain_name,
                user_domain_id=self.user_domain_id,
                project_id=self.project_id or self.tenant_id,
                project_name=self.project_name,
                project_domain_name=self.project_domain_name,
                project_domain_id=self.project_domain_id,
                region_name=self.region_name)
        elif v2_auth_url:
            keystone_client = ks_client.Client(
                session=ks_session,
                version=(2, 0),
                auth_url=v2_auth_url,
                username=self.username,
                password=self.password,
                tenant_id=self.tenant_id,
                tenant_name=self.tenant_name,
                region_name=self.region_name,
                cert=self.cert,
                use_keyring=self.use_keyring,
                force_new_token=self.force_new_token,
                stale_duration=self.cached_token_lifetime)
        else:
            raise exceptions.CommandError(
                'Unable to determine the Keystone version to authenticate '
                'with using the given auth_url.')
        keystone_client.authenticate()
        return keystone_client
